@@ -7,7 +7,6 @@
 // much better "put pkg X at folder Y" abstraction.  Oh well,
 // whatever.  Perfect enemy of the good, and all that.
 
-var url = require("url")
 var fs = require("fs")
 var asyncMap = require("slide").asyncMap
 var path = require("path")
@@ -16,6 +15,7 @@ var semver = require("semver")
 var rm = require("./utils/gently-rm.js")
 var log = require("npmlog")
 var npm = require("./npm.js")
+var mapToRegistry = require("./utils/map-to-registry.js")
 
 module.exports = dedupe
 
@@ -61,7 +61,7 @@ function dedupe_ (dir, filter, unavoidable, dryrun, silent, cb) {
       Object.keys(obj.children).forEach(function (k) {
         U(obj.children[k])
       })
-    })
+    })(data)
 
     // then collect them up and figure out who needs them
     ;(function C (obj) {
@@ -240,13 +240,19 @@ function findVersions (npm, summary, cb) {
     var versions = data.versions
 
     var ranges = data.ranges
-    var uri = url.resolve(npm.config.get("registry"), name)
-    npm.registry.get(uri, null, function (er, data) {
+    mapToRegistry(name, npm.config, function (er, uri, auth) {
+      if (er) return cb(er)
+
+      npm.registry.get(uri, { auth : auth }, next)
+    })
+
+    function next (er, data) {
       var regVersions = er ? [] : Object.keys(data.versions)
       var locMatch = bestMatch(versions, ranges)
-      var regMatch;
       var tag = npm.config.get("tag")
       var distTag = data["dist-tags"] && data["dist-tags"][tag]
+
+      var regMatch
       if (distTag && data.versions[distTag] && matches(distTag, ranges)) {
         regMatch = distTag
       } else {
@@ -254,7 +260,7 @@ function findVersions (npm, summary, cb) {
       }
 
       cb(null, [[name, has, loc, locMatch, regMatch, locs]])
-    })
+    }
   }, cb)
 }
 
