@@ -79,12 +79,13 @@ static void uv__fast_poll_submit_poll_req(uv_loop_t* loop, uv_poll_t* handle) {
     handle->mask_events_2 = handle->events;
   } else if (handle->submitted_events_2 == 0) {
     req = &handle->poll_req_2;
-    afd_poll_info = &handle->afd_poll_info_2;
+    afd_poll_info = &handle->afd_poll_info_2.afd_poll_info_ptr[0];
     handle->submitted_events_2 = handle->events;
     handle->mask_events_1 = handle->events;
     handle->mask_events_2 = 0;
   } else {
     assert(0);
+    return;
   }
 
   /* Setting Exclusive to TRUE makes the other poll request return if there */
@@ -118,25 +119,25 @@ static void uv__fast_poll_submit_poll_req(uv_loop_t* loop, uv_poll_t* handle) {
 
 
 static int uv__fast_poll_cancel_poll_req(uv_loop_t* loop, uv_poll_t* handle) {
-  AFD_POLL_INFO afd_poll_info;
+  AFD_POLL_INFO* afd_poll_info;
   DWORD result;
 
-  afd_poll_info.Exclusive = TRUE;
-  afd_poll_info.NumberOfHandles = 1;
-  afd_poll_info.Timeout.QuadPart = INT64_MAX;
-  afd_poll_info.Handles[0].Handle = (HANDLE) handle->socket;
-  afd_poll_info.Handles[0].Status = 0;
-  afd_poll_info.Handles[0].Events = AFD_POLL_ALL;
+  afd_poll_info = &handle->afd_poll_info_2.afd_poll_info_ptr[1];
+  afd_poll_info->Exclusive = TRUE;
+  afd_poll_info->NumberOfHandles = 1;
+  afd_poll_info->Timeout.QuadPart = INT64_MAX;
+  afd_poll_info->Handles[0].Handle = (HANDLE) handle->socket;
+  afd_poll_info->Handles[0].Status = 0;
+  afd_poll_info->Handles[0].Events = AFD_POLL_ALL;
 
   result = uv_msafd_poll(handle->socket,
-                         &afd_poll_info,
+                         afd_poll_info,
                          uv__get_overlapped_dummy());
 
   if (result == SOCKET_ERROR) {
     DWORD error = WSAGetLastError();
     if (error != WSA_IO_PENDING) {
-      uv__set_sys_error(loop, WSAGetLastError());
-      return -1;
+      return WSAGetLastError();
     }
   }
 
@@ -154,11 +155,12 @@ static void uv__fast_poll_process_poll_req(uv_loop_t* loop, uv_poll_t* handle,
     handle->submitted_events_1 = 0;
     mask_events = handle->mask_events_1;
   } else if (req == &handle->poll_req_2) {
-    afd_poll_info = &handle->afd_poll_info_2;
+    afd_poll_info = &handle->afd_poll_info_2.afd_poll_info_ptr[0];
     handle->submitted_events_2 = 0;
     mask_events = handle->mask_events_2;
   } else {
     assert(0);
+    return;
   }
 
   /* Report an error unless the select was just interrupted. */
@@ -166,8 +168,7 @@ static void uv__fast_poll_process_poll_req(uv_loop_t* loop, uv_poll_t* handle,
     DWORD error = GET_REQ_SOCK_ERROR(req);
     if (error != WSAEINTR && handle->events != 0) {
       handle->events = 0; /* Stop the watcher */
-      uv__set_sys_error(loop, error);
-      handle->poll_cb(handle, -1, 0);
+      handle->poll_cb(handle, uv_translate_sys_error(error), 0);
     }
 
   } else if (afd_poll_info->NumberOfHandles >= 1) {
@@ -229,17 +230,18 @@ static int uv__fast_poll_set(uv_loop_t* loop, uv_poll_t* handle, int events) {
 }
 
 
-static void uv__fast_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
+static int uv__fast_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
   handle->events = 0;
   uv__handle_closing(handle);
 
   if (handle->submitted_events_1 == 0 &&
       handle->submitted_events_2 == 0) {
     uv_want_endgame(loop, (uv_handle_t*) handle);
+    return 0;
   } else {
     /* Cancel outstanding poll requests by executing another, unique poll */
     /* request that forces the outstanding ones to return. */
-    uv__fast_poll_cancel_poll_req(loop, handle);
+    return uv__fast_poll_cancel_poll_req(loop, handle);
   }
 }
 
@@ -283,7 +285,7 @@ static SOCKET uv__fast_poll_get_peer_socket(uv_loop_t* loop,
   SOCKET peer_socket;
 
   index = -1;
-  for (i = 0; i < ARRAY_SIZE(uv_msafd_provider_ids); i++) {
+  for (i = 0; (size_t) i < ARRAY_SIZE(uv_msafd_provider_ids); i++) {
     if (memcmp((void*) &protocol_info->ProviderId,
                (void*) &uv_msafd_provider_ids[i],
                sizeof protocol_info->ProviderId) == 0) {
@@ -394,6 +396,7 @@ static void uv__slow_poll_submit_poll_req(uv_loop_t* loop, uv_poll_t* handle) {
     handle->mask_events_2 = 0;
   } else {
     assert(0);
+    return;
   }
 
   if (!QueueUserWorkItem(uv__slow_poll_thread_proc,
@@ -410,6 +413,8 @@ static void uv__slow_poll_submit_poll_req(uv_loop_t* loop, uv_poll_t* handle) {
 static void uv__slow_poll_process_poll_req(uv_loop_t* loop, uv_poll_t* handle,
     uv_req_t* req) {
   unsigned char mask_events;
+  int err;
+
   if (req == &handle->poll_req_1) {
     handle->submitted_events_1 = 0;
     mask_events = handle->mask_events_1;
@@ -418,14 +423,15 @@ static void uv__slow_poll_process_poll_req(uv_loop_t* loop, uv_poll_t* handle,
     mask_events = handle->mask_events_2;
   } else {
     assert(0);
+    return;
   }
 
   if (!REQ_SUCCESS(req)) {
     /* Error. */
     if (handle->events != 0) {
+      err = GET_REQ_ERROR(req);
       handle->events = 0; /* Stop the watcher */
-      uv__set_sys_error(loop, GET_REQ_ERROR(req));
-      handle->poll_cb(handle, -1, 0);
+      handle->poll_cb(handle, uv_translate_sys_error(err), 0);
     }
   } else {
     /* Got some events. */
@@ -468,7 +474,7 @@ static int uv__slow_poll_set(uv_loop_t* loop, uv_poll_t* handle, int events) {
 }
 
 
-static void uv__slow_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
+static int uv__slow_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
   handle->events = 0;
   uv__handle_closing(handle);
 
@@ -476,11 +482,13 @@ static void uv__slow_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
       handle->submitted_events_2 == 0) {
     uv_want_endgame(loop, (uv_handle_t*) handle);
   }
+
+  return 0;
 }
 
 
 int uv_poll_init(uv_loop_t* loop, uv_poll_t* handle, int fd) {
-  return uv_poll_init_socket(loop, handle, (SOCKET) _get_osfhandle(fd));
+  return uv_poll_init_socket(loop, handle, (SOCKET) uv__get_osfhandle(fd));
 }
 
 
@@ -523,8 +531,7 @@ int uv_poll_init_socket(uv_loop_t* loop, uv_poll_t* handle,
                  SO_PROTOCOL_INFOW,
                  (char*) &protocol_info,
                  &len) != 0) {
-    uv__set_sys_error(loop, WSAGetLastError());
-    return -1;
+    return uv_translate_sys_error(WSAGetLastError());
   }
 
   /* Get the peer socket that is needed to enable fast poll. If the returned */
@@ -540,7 +547,7 @@ int uv_poll_init_socket(uv_loop_t* loop, uv_poll_t* handle,
     handle->flags |= UV_HANDLE_POLL_SLOW;
   }
 
-  /* Intialize 2 poll reqs. */
+  /* Initialize 2 poll reqs. */
   handle->submitted_events_1 = 0;
   uv_req_init(loop, (uv_req_t*) &(handle->poll_req_1));
   handle->poll_req_1.type = UV_POLL_REQ;
@@ -551,17 +558,26 @@ int uv_poll_init_socket(uv_loop_t* loop, uv_poll_t* handle,
   handle->poll_req_2.type = UV_POLL_REQ;
   handle->poll_req_2.data = handle;
 
+  handle->afd_poll_info_2.afd_poll_info_ptr = malloc(sizeof(*handle->afd_poll_info_2.afd_poll_info_ptr) * 2);
+  if (handle->afd_poll_info_2.afd_poll_info_ptr == NULL) {
+    return UV_ENOMEM;
+  }
+
   return 0;
 }
 
 
 int uv_poll_start(uv_poll_t* handle, int events, uv_poll_cb cb) {
+  int err;
+
   if (!(handle->flags & UV_HANDLE_POLL_SLOW)) {
-    if (uv__fast_poll_set(handle->loop, handle, events) < 0)
-      return -1;
+    err = uv__fast_poll_set(handle->loop, handle, events);
   } else {
-    if (uv__slow_poll_set(handle->loop, handle, events) < 0)
-      return -1;
+    err = uv__slow_poll_set(handle->loop, handle, events);
+  }
+
+  if (err) {
+    return uv_translate_sys_error(err);
   }
 
   handle->poll_cb = cb;
@@ -571,11 +587,15 @@ int uv_poll_start(uv_poll_t* handle, int events, uv_poll_cb cb) {
 
 
 int uv_poll_stop(uv_poll_t* handle) {
+  int err;
+
   if (!(handle->flags & UV_HANDLE_POLL_SLOW)) {
-    return uv__fast_poll_set(handle->loop, handle, 0);
+    err = uv__fast_poll_set(handle->loop, handle, 0);
   } else {
-    return uv__slow_poll_set(handle->loop, handle, 0);
+    err = uv__slow_poll_set(handle->loop, handle, 0);
   }
+
+  return uv_translate_sys_error(err);
 }
 
 
@@ -588,11 +608,11 @@ void uv_process_poll_req(uv_loop_t* loop, uv_poll_t* handle, uv_req_t* req) {
 }
 
 
-void uv_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
+int uv_poll_close(uv_loop_t* loop, uv_poll_t* handle) {
   if (!(handle->flags & UV_HANDLE_POLL_SLOW)) {
-    uv__fast_poll_close(loop, handle);
+    return uv__fast_poll_close(loop, handle);
   } else {
-    uv__slow_poll_close(loop, handle);
+    return uv__slow_poll_close(loop, handle);
   }
 }
 
@@ -604,5 +624,9 @@ void uv_poll_endgame(uv_loop_t* loop, uv_poll_t* handle) {
   assert(handle->submitted_events_1 == 0);
   assert(handle->submitted_events_2 == 0);
 
+  if (handle->afd_poll_info_2.afd_poll_info_ptr) {
+    free(handle->afd_poll_info_2.afd_poll_info_ptr);
+    handle->afd_poll_info_2.afd_poll_info_ptr = NULL;
+  }
   uv__handle_close(handle);
 }
