@@ -25,6 +25,7 @@ var util = require('util'),
     vm = require('vm'),
     repl = require('repl'),
     inherits = util.inherits,
+    assert = require('assert'),
     spawn = require('child_process').spawn;
 
 exports.start = function(argv, stdin, stdout) {
@@ -164,7 +165,6 @@ function Client() {
 
   this.currentFrame = NO_FRAME;
   this.currentSourceLine = -1;
-  this.currentSource = null;
   this.handles = {};
   this.scripts = {};
   this.breakpoints = [];
@@ -182,7 +182,7 @@ exports.Client = Client;
 
 
 Client.prototype._addHandle = function(desc) {
-  if (typeof desc != 'object' || typeof desc.handle != 'number') {
+  if (!util.isObject(desc) || !util.isNumber(desc.handle)) {
     return;
   }
 
@@ -249,6 +249,10 @@ Client.prototype._onResponse = function(res) {
     this._removeScript(res.body.body.script);
     handled = true;
 
+  } else if (res.body && res.body.event === 'compileError') {
+    // This event is not used anywhere right now, perhaps somewhere in the
+    // future?
+    handled = true;
   }
 
   if (cb) {
@@ -296,7 +300,7 @@ Client.prototype.reqLookup = function(refs, cb) {
   this.req(req, function(err, res) {
     if (err) return cb(err);
     for (var ref in res) {
-      if (typeof res[ref] == 'object') {
+      if (util.isObject(res[ref])) {
         self._addHandle(res[ref]);
       }
     }
@@ -366,7 +370,7 @@ Client.prototype.reqEval = function(expression, cb) {
 };
 
 
-// Finds the first scope in the array in which the epxression evals.
+// Finds the first scope in the array in which the expression evals.
 Client.prototype._reqFramesEval = function(expression, evalFrames, cb) {
   if (evalFrames.length == 0) {
     // Just eval in global scope.
@@ -559,8 +563,7 @@ Client.prototype.mirrorObject = function(handle, depth, cb) {
         }
 
 
-        if (Array.isArray(mirror) &&
-            typeof prop.name != 'number') {
+        if (util.isArray(mirror) && !util.isNumber(prop.name)) {
           // Skip the 'length' property.
           return;
         }
@@ -593,7 +596,7 @@ Client.prototype.mirrorObject = function(handle, depth, cb) {
     val = function() {};
   } else if (handle.type === 'null') {
     val = null;
-  } else if (handle.value !== undefined) {
+  } else if (!util.isUndefined(handle.value)) {
     val = handle.value;
   } else if (handle.type === 'undefined') {
     val = undefined;
@@ -740,8 +743,7 @@ function SourceInfo(body) {
 // This class is the repl-enabled debugger interface which is invoked on
 // "node debug"
 function Interface(stdin, stdout, args) {
-  var self = this,
-      child;
+  var self = this;
 
   this.stdin = stdin;
   this.stdout = stdout;
@@ -857,13 +859,14 @@ function Interface(stdin, stdout, args) {
 
 
 Interface.prototype.pause = function() {
-  if (this.killed || this.paused++ > 0) return false;
+  if (this.killed || this.paused++ > 0) return this;
   this.repl.rli.pause();
   this.stdin.pause();
+  return this;
 };
 
 Interface.prototype.resume = function(silent) {
-  if (this.killed || this.paused === 0 || --this.paused !== 0) return false;
+  if (this.killed || this.paused === 0 || --this.paused !== 0) return this;
   this.repl.rli.resume();
   if (silent !== true) {
     this.repl.displayPrompt();
@@ -874,6 +877,7 @@ Interface.prototype.resume = function(silent) {
     this.waiting();
     this.waiting = null;
   }
+  return this;
 };
 
 
@@ -892,7 +896,7 @@ Interface.prototype.print = function(text, oneline) {
   if (this.killed) return;
   this.clearline();
 
-  this.stdout.write(typeof text === 'string' ? text : util.inspect(text));
+  this.stdout.write(util.isString(text) ? text : util.inspect(text));
 
   if (oneline !== true) {
     this.stdout.write('\n');
@@ -961,8 +965,8 @@ Interface.prototype.controlEval = function(code, context, filename, callback) {
   try {
     // Repeat last command if empty line are going to be evaluated
     if (this.repl.rli.history && this.repl.rli.history.length > 0) {
-      if (code === '(\n)') {
-        code = '(' + this.repl.rli.history[0] + '\n)';
+      if (code === '\n') {
+        code = this.repl.rli.history[0] + '\n';
       }
     }
 
@@ -1018,28 +1022,13 @@ Interface.prototype.debugEval = function(code, context, filename, callback) {
 
 // Utils
 
-// Returns number of digits (+1)
-function intChars(n) {
-  // TODO dumb:
-  if (n < 50) {
-    return 3;
-  } else if (n < 950) {
-    return 4;
-  } else if (n < 9950) {
-    return 5;
-  } else {
-    return 6;
-  }
-}
-
 // Adds spaces and prefix to number
-function leftPad(n, prefix) {
+// maxN is a maximum number we should have space for
+function leftPad(n, prefix, maxN) {
   var s = n.toString(),
-      nchars = intChars(n),
+      nchars = Math.max(2, String(maxN).length) + 1,
       nspaces = nchars - s.length - 1;
 
-  prefix || (prefix = ' ');
-
   for (var i = 0; i < nspaces; i++) {
     prefix += ' ';
   }
@@ -1130,8 +1119,9 @@ Interface.prototype.list = function(delta) {
 
       var current = lineno == 1 + client.currentSourceLine,
           breakpoint = client.breakpoints.some(function(bp) {
-            return bp.script === client.currentScript &&
-                   bp.line == lineno;
+            return (bp.scriptReq === client.currentScript ||
+                    bp.script === client.currentScript) &&
+                    bp.line == lineno;
           });
 
       if (lineno == 1) {
@@ -1153,7 +1143,14 @@ Interface.prototype.list = function(delta) {
         line = lines[i];
       }
 
-      self.print(leftPad(lineno, breakpoint && '*') + ' ' + line);
+      var prefixChar = ' ';
+      if (current) {
+        prefixChar = '>';
+      } else if (breakpoint) {
+        prefixChar = '*';
+      }
+
+      self.print(leftPad(lineno, prefixChar, to) + ' ' + line);
     }
     self.resume();
   });
@@ -1214,7 +1211,7 @@ Interface.prototype.scripts = function() {
   this.pause();
   for (var id in client.scripts) {
     var script = client.scripts[id];
-    if (typeof script == 'object' && script.name) {
+    if (util.isObject(script) && script.name) {
       if (displayNatives ||
           script.name == client.currentScript ||
           !script.isNative) {
@@ -1315,7 +1312,8 @@ Interface.prototype.watchers = function() {
       if (verbose) self.print('Watchers:');
 
       self._watchers.forEach(function(watcher, i) {
-        self.print(leftPad(i, ' ') + ': ' + watcher + ' = ' +
+        self.print(leftPad(i, ' ', self._watchers.length - 1) +
+                   ': ' + watcher + ' = ' +
                    JSON.stringify(values[i]));
       });
 
@@ -1351,17 +1349,23 @@ Interface.prototype.setBreakpoint = function(script, line,
       ambiguous;
 
   // setBreakpoint() should insert breakpoint on current line
-  if (script === undefined) {
+  if (util.isUndefined(script)) {
     script = this.client.currentScript;
     line = this.client.currentSourceLine + 1;
   }
 
   // setBreakpoint(line-number) should insert breakpoint in current script
-  if (line === undefined && typeof script === 'number') {
+  if (util.isUndefined(line) && util.isNumber(script)) {
     line = script;
     script = this.client.currentScript;
   }
 
+  if (script === undefined) {
+    this.print('Cannot determine the current script, ' +
+        'make sure the debugged process is paused.');
+    return;
+  }
+
   if (/\(\)$/.test(script)) {
     // setBreakpoint('functionname()');
     var req = {
@@ -1373,14 +1377,18 @@ Interface.prototype.setBreakpoint = function(script, line,
     // setBreakpoint('scriptname')
     if (script != +script && !this.client.scripts[script]) {
       var scripts = this.client.scripts;
-      Object.keys(scripts).forEach(function(id) {
-        if (scripts[id] && scripts[id].name.indexOf(script) !== -1) {
+      var keys = Object.keys(scripts);
+      for (var v = 0; v < keys.length; v++) {
+        var id = keys[v];
+        if (scripts[id] &&
+            scripts[id].name &&
+            scripts[id].name.indexOf(script) !== -1) {
           if (scriptId) {
             ambiguous = true;
           }
           scriptId = id;
         }
-      });
+      }
     } else {
       scriptId = script;
     }
@@ -1452,7 +1460,7 @@ Interface.prototype.clearBreakpoint = function(script, line) {
     if (bp.scriptId === script ||
         bp.scriptReq === script ||
         (bp.script && bp.script.indexOf(script) !== -1)) {
-      if (index !== undefined) {
+      if (!util.isUndefined(index)) {
         ambiguous = true;
       }
       if (bp.line === line) {
@@ -1465,7 +1473,7 @@ Interface.prototype.clearBreakpoint = function(script, line) {
 
   if (ambiguous) return this.error('Script name is ambiguous');
 
-  if (breakpoint === undefined) {
+  if (util.isUndefined(breakpoint)) {
     return this.error('Script : ' + script + ' not found');
   }
 
@@ -1562,7 +1570,6 @@ Interface.prototype.repl = function() {
   this.history.control = this.repl.rli.history;
   this.repl.rli.history = this.history.debug;
 
-  this.repl.prompt = '> ';
   this.repl.rli.setPrompt('> ');
   this.repl.displayPrompt();
 };
@@ -1578,7 +1585,6 @@ Interface.prototype.exitRepl = function() {
   this.repl.rli.history = this.history.control;
 
   this.repl.context = this.context;
-  this.repl.prompt = 'debug> ';
   this.repl.rli.setPrompt('debug> ');
   this.repl.displayPrompt();
 };
@@ -1617,6 +1623,7 @@ Interface.prototype.trySpawn = function(cb) {
       childArgs = this.args;
 
   this.killChild();
+  assert(!this.child);
 
   if (this.args.length === 2) {
     var match = this.args[1].match(/^([^:]+):(\d+)$/);
@@ -1652,12 +1659,10 @@ Interface.prototype.trySpawn = function(cb) {
     }
   }
 
-  if (!this.child) {
-    this.child = spawn(process.execPath, childArgs);
+  this.child = spawn(process.execPath, childArgs);
 
-    this.child.stdout.on('data', this.childPrint.bind(this));
-    this.child.stderr.on('data', this.childPrint.bind(this));
-  }
+  this.child.stdout.on('data', this.childPrint.bind(this));
+  this.child.stderr.on('data', this.childPrint.bind(this));
 
   this.pause();
 
@@ -1714,8 +1719,10 @@ Interface.prototype.trySpawn = function(cb) {
     client.connect(port, host);
   }
 
-  setTimeout(function() {
-    self.print('connecting..', true);
-    attemptConnect();
-  }, 50);
+  this.child.stderr.once('data', function() {
+    setImmediate(function() {
+      self.print('connecting to port ' + port + '..', true);
+      attemptConnect();
+    });
+  });
 };
