@@ -23,11 +23,17 @@ var StringDecoder = require('string_decoder').StringDecoder;
 var EventEmitter = require('events').EventEmitter;
 var net = require('net');
 var dgram = require('dgram');
-var Process = process.binding('process_wrap').Process;
 var assert = require('assert');
 var util = require('util');
-var constants; // if (!constants) constants = process.binding('constants');
 
+var Process = process.binding('process_wrap').Process;
+var WriteWrap = process.binding('stream_wrap').WriteWrap;
+var uv = process.binding('uv');
+
+var spawn_sync; // Lazy-loaded process.binding('spawn_sync')
+var constants;  // Lazy-loaded process.binding('constants')
+
+var errnoException = util._errnoException;
 var handleWraps = {};
 
 function handleWrapGetter(name, callback) {
@@ -35,7 +41,7 @@ function handleWrapGetter(name, callback) {
 
   Object.defineProperty(handleWraps, name, {
     get: function() {
-      if (cons !== undefined) return cons;
+      if (!util.isUndefined(cons)) return cons;
       return cons = callback();
     }
   });
@@ -100,8 +106,6 @@ var handleConversion = {
     },
 
     got: function(message, handle, emit) {
-      var self = this;
-
       var server = new net.Server();
       server.listen(handle, function() {
         emit(server);
@@ -199,8 +203,6 @@ var handleConversion = {
 function SocketListSend(slave, key) {
   EventEmitter.call(this);
 
-  var self = this;
-
   this.key = key;
   this.slave = slave;
 }
@@ -312,9 +314,9 @@ function getSocketList(type, slave, key) {
 var INTERNAL_PREFIX = 'NODE_';
 function handleMessage(target, message, handle) {
   var eventName = 'message';
-  if (message !== null &&
-      typeof message === 'object' &&
-      typeof message.cmd === 'string' &&
+  if (!util.isNull(message) &&
+      util.isObject(message) &&
+      util.isString(message.cmd) &&
       message.cmd.length > INTERNAL_PREFIX.length &&
       message.cmd.slice(0, INTERNAL_PREFIX.length) === INTERNAL_PREFIX) {
     eventName = 'internalMessage';
@@ -329,9 +331,10 @@ function setupChannel(target, channel) {
   var decoder = new StringDecoder('utf8');
   var jsonBuffer = '';
   channel.buffering = false;
-  channel.onread = function(pool, offset, length, recvHandle) {
+  channel.onread = function(nread, pool, recvHandle) {
+    // TODO(bnoordhuis) Check that nread > 0.
     if (pool) {
-      jsonBuffer += decoder.write(pool.slice(offset, offset + length));
+      jsonBuffer += decoder.write(pool);
 
       var i, start = 0;
 
@@ -369,19 +372,28 @@ function setupChannel(target, channel) {
   target.on('internalMessage', function(message, handle) {
     // Once acknowledged - continue sending handles.
     if (message.cmd === 'NODE_HANDLE_ACK') {
-      assert(Array.isArray(target._handleQueue));
+      assert(util.isArray(target._handleQueue));
       var queue = target._handleQueue;
       target._handleQueue = null;
+
       queue.forEach(function(args) {
-        target.send(args.message, args.handle);
+        target._send(args.message, args.handle, false);
       });
+
+      // Process a pending disconnect (if any).
+      if (!target.connected && target._channel && !target._handleQueue)
+        target._disconnect();
+
       return;
     }
 
     if (message.cmd !== 'NODE_HANDLE') return;
 
-    // Acknowledge handle receival.
-    target.send({ cmd: 'NODE_HANDLE_ACK' });
+    // Acknowledge handle receival. Don't emit error events (for example if
+    // the other side has disconnected) because this call to send() is not
+    // initiated by the user and it shouldn't be fatal to be unable to ACK
+    // a message.
+    target._send({ cmd: 'NODE_HANDLE_ACK' }, null, true);
 
     var obj = handleConversion[message.type];
 
@@ -398,14 +410,17 @@ function setupChannel(target, channel) {
   });
 
   target.send = function(message, handle) {
-    if (typeof message === 'undefined') {
-      throw new TypeError('message cannot be undefined');
-    }
-
-    if (!this.connected) {
+    if (!this.connected)
       this.emit('error', new Error('channel closed'));
-      return;
-    }
+    else
+      this._send(message, handle, false);
+  };
+
+  target._send = function(message, handle, swallowErrors) {
+    assert(this.connected || this._channel);
+
+    if (util.isUndefined(message))
+      throw new TypeError('message cannot be undefined');
 
     // package messages with a handle object
     if (handle) {
@@ -440,7 +455,8 @@ function setupChannel(target, channel) {
       var obj = handleConversion[message.type];
 
       // convert TCP object to native handle object
-      handle = handleConversion[message.type].send.apply(target, arguments);
+      handle =
+          handleConversion[message.type].send.call(target, message, handle);
 
       // If handle was sent twice, or it is impossible to get native handle
       // out of it - just send a text without the handle.
@@ -458,37 +474,57 @@ function setupChannel(target, channel) {
       return;
     }
 
+    var req = new WriteWrap();
+    req.oncomplete = nop;
     var string = JSON.stringify(message) + '\n';
-    var writeReq = channel.writeUtf8String(string, handle);
+    var err = channel.writeUtf8String(req, string, handle);
 
-    if (!writeReq) {
-      var er = errnoException(process._errno,
-                              'write',
-                              'cannot write to IPC channel.');
-      this.emit('error', er);
+    if (err) {
+      if (!swallowErrors)
+        this.emit('error', errnoException(err, 'write'));
     } else if (handle && !this._handleQueue) {
       this._handleQueue = [];
     }
 
     if (obj && obj.postSend) {
-      writeReq.oncomplete = obj.postSend.bind(null, handle);
-    } else {
-      writeReq.oncomplete = nop;
+      req.oncomplete = obj.postSend.bind(null, handle);
     }
 
     /* If the master is > 2 read() calls behind, please stop sending. */
     return channel.writeQueueSize < (65536 * 2);
   };
 
+  // connected will be set to false immediately when a disconnect() is
+  // requested, even though the channel might still be alive internally to
+  // process queued messages. The three states are distinguished as follows:
+  // - disconnect() never requested: _channel is not null and connected
+  //   is true
+  // - disconnect() requested, messages in the queue: _channel is not null
+  //   and connected is false
+  // - disconnect() requested, channel actually disconnected: _channel is
+  //   null and connected is false
   target.connected = true;
+
   target.disconnect = function() {
     if (!this.connected) {
       this.emit('error', new Error('IPC channel is already disconnected'));
       return;
     }
 
-    // do not allow messages to be written
+    // Do not allow any new messages to be written.
     this.connected = false;
+
+    // If there are no queued messages, disconnect immediately. Otherwise,
+    // postpone the disconnect so that it happens internally after the
+    // queue is flushed.
+    if (!this._handleQueue)
+      this._disconnect();
+  };
+
+  target._disconnect = function() {
+    assert(this._channel);
+
+    // This marks the fact that the channel is actually disconnected.
     this._channel = null;
 
     var fired = false;
@@ -508,7 +544,7 @@ function setupChannel(target, channel) {
       return;
     }
 
-    finish();
+    process.nextTick(finish);
   };
 
   channel.readStart();
@@ -521,9 +557,11 @@ exports.fork = function(modulePath /*, args, options*/) {
 
   // Get options and args arguments.
   var options, args, execArgv;
-  if (Array.isArray(arguments[1])) {
+  if (util.isArray(arguments[1])) {
     args = arguments[1];
     options = util._extend({}, arguments[2]);
+  } else if (arguments[1] && typeof arguments[1] !== 'object') {
+    throw new TypeError('Incorrect value of args option');
   } else {
     args = [];
     options = util._extend({}, arguments[1]);
@@ -563,10 +601,10 @@ exports._forkChild = function(fd) {
 };
 
 
-exports.exec = function(command /*, options, callback */) {
+function normalizeExecArgs(command /*, options, callback */) {
   var file, args, options, callback;
 
-  if (typeof arguments[1] === 'function') {
+  if (util.isFunction(arguments[1])) {
     options = undefined;
     callback = arguments[1];
   } else {
@@ -575,7 +613,7 @@ exports.exec = function(command /*, options, callback */) {
   }
 
   if (process.platform === 'win32') {
-    file = 'cmd.exe';
+    file = process.env.comspec || 'cmd.exe';
     args = ['/s', '/c', '"' + command + '"'];
     // Make a shallow copy before patching so we don't clobber the user's
     // options object.
@@ -585,12 +623,31 @@ exports.exec = function(command /*, options, callback */) {
     file = '/bin/sh';
     args = ['-c', command];
   }
-  return exports.execFile(file, args, options, callback);
+
+  if (options && options.shell)
+    file = options.shell;
+
+  return {
+    cmd: command,
+    file: file,
+    args: args,
+    options: options,
+    callback: callback
+  };
+}
+
+
+exports.exec = function(command /*, options, callback */) {
+  var opts = normalizeExecArgs.apply(null, arguments);
+  return exports.execFile(opts.file,
+                          opts.args,
+                          opts.options,
+                          opts.callback);
 };
 
 
 exports.execFile = function(file /* args, options, callback */) {
-  var args, optionArg, callback;
+  var args = [], callback;
   var options = {
     encoding: 'utf8',
     timeout: 0,
@@ -600,33 +657,55 @@ exports.execFile = function(file /* args, options, callback */) {
     env: null
   };
 
-  // Parse the parameters.
-
-  if (typeof arguments[arguments.length - 1] === 'function') {
-    callback = arguments[arguments.length - 1];
+  // Parse the optional positional parameters.
+  var pos = 1;
+  if (pos < arguments.length && Array.isArray(arguments[pos])) {
+    args = arguments[pos++];
+  } else if (pos < arguments.length && arguments[pos] == null) {
+    pos++;
   }
 
-  if (Array.isArray(arguments[1])) {
-    args = arguments[1];
-    options = util._extend(options, arguments[2]);
-  } else {
-    args = [];
-    options = util._extend(options, arguments[1]);
+  if (pos < arguments.length && util.isObject(arguments[pos])) {
+    options = util._extend(options, arguments[pos++]);
+  } else if (pos < arguments.length && arguments[pos] == null) {
+    pos++;
+  }
+
+  if (pos < arguments.length && util.isFunction(arguments[pos])) {
+    callback = arguments[pos++];
+  }
+
+  if (pos === 1 && arguments.length > 1) {
+    throw new TypeError('Incorrect value of args option');
   }
 
   var child = spawn(file, args, {
     cwd: options.cwd,
     env: options.env,
+    gid: options.gid,
+    uid: options.uid,
     windowsVerbatimArguments: !!options.windowsVerbatimArguments
   });
 
-  var stdout = '';
-  var stderr = '';
+  var encoding;
+  var _stdout;
+  var _stderr;
+  if (options.encoding !== 'buffer' && Buffer.isEncoding(options.encoding)) {
+    encoding = options.encoding;
+    _stdout = '';
+    _stderr = '';
+  } else {
+    _stdout = [];
+    _stderr = [];
+    encoding = null;
+  }
+  var stdoutLen = 0;
+  var stderrLen = 0;
   var killed = false;
   var exited = false;
   var timeoutId;
 
-  var err;
+  var ex = null;
 
   function exithandler(code, signal) {
     if (exited) return;
@@ -639,21 +718,41 @@ exports.execFile = function(file /* args, options, callback */) {
 
     if (!callback) return;
 
-    if (err) {
-      callback(err, stdout, stderr);
+    // merge chunks
+    var stdout;
+    var stderr;
+    if (!encoding) {
+      stdout = Buffer.concat(_stdout);
+      stderr = Buffer.concat(_stderr);
+    } else {
+      stdout = _stdout;
+      stderr = _stderr;
+    }
+
+    if (ex) {
+      // Will be handled later
     } else if (code === 0 && signal === null) {
       callback(null, stdout, stderr);
-    } else {
-      var e = new Error('Command failed: ' + stderr);
-      e.killed = child.killed || killed;
-      e.code = code;
-      e.signal = signal;
-      callback(e, stdout, stderr);
+      return;
     }
+
+    var cmd = file;
+    if (args.length !== 0)
+      cmd += ' ' + args.join(' ');
+
+    if (!ex) {
+      ex = new Error('Command failed: ' + cmd + '\n' + stderr);
+      ex.killed = child.killed || killed;
+      ex.code = code < 0 ? uv.errname(code) : code;
+      ex.signal = signal;
+    }
+
+    ex.cmd = cmd;
+    callback(ex, stdout, stderr);
   }
 
   function errorhandler(e) {
-    err = e;
+    ex = e;
     child.stdout.destroy();
     child.stderr.destroy();
     exithandler();
@@ -667,7 +766,7 @@ exports.execFile = function(file /* args, options, callback */) {
     try {
       child.kill(options.killSignal);
     } catch (e) {
-      err = e;
+      ex = e;
       exithandler();
     }
   }
@@ -679,67 +778,223 @@ exports.execFile = function(file /* args, options, callback */) {
     }, options.timeout);
   }
 
-  child.stdout.setEncoding(options.encoding);
-  child.stderr.setEncoding(options.encoding);
-
   child.stdout.addListener('data', function(chunk) {
-    stdout += chunk;
-    if (stdout.length > options.maxBuffer) {
-      err = new Error('stdout maxBuffer exceeded.');
+    stdoutLen += chunk.length;
+
+    if (stdoutLen > options.maxBuffer) {
+      ex = new Error('stdout maxBuffer exceeded.');
       kill();
+    } else {
+      if (!encoding)
+        _stdout.push(chunk);
+      else
+        _stdout += chunk;
     }
   });
 
   child.stderr.addListener('data', function(chunk) {
-    stderr += chunk;
-    if (stderr.length > options.maxBuffer) {
-      err = new Error('stderr maxBuffer exceeded.');
+    stderrLen += chunk.length;
+
+    if (stderrLen > options.maxBuffer) {
+      ex = new Error('stderr maxBuffer exceeded.');
       kill();
+    } else {
+      if (!encoding)
+        _stderr.push(chunk);
+      else
+        _stderr += chunk;
     }
   });
 
+  if (encoding) {
+    child.stderr.setEncoding(encoding);
+    child.stdout.setEncoding(encoding);
+  }
+
   child.addListener('close', exithandler);
   child.addListener('error', errorhandler);
 
   return child;
 };
 
+var _deprecatedCustomFds = util.deprecate(function(options) {
+  options.stdio = options.customFds.map(function(fd) {
+    return fd === -1 ? 'pipe' : fd;
+  });
+}, 'child_process: customFds option is deprecated, use stdio instead.');
 
-var spawn = exports.spawn = function(file /*, args, options*/) {
+function _convertCustomFds(options) {
+  if (options && options.customFds && !options.stdio) {
+    _deprecatedCustomFds(options);
+  }
+}
+
+
+function _validateStdio(stdio, sync) {
+  var ipc,
+      ipcFd;
+
+  // Replace shortcut with an array
+  if (util.isString(stdio)) {
+    switch (stdio) {
+      case 'ignore': stdio = ['ignore', 'ignore', 'ignore']; break;
+      case 'pipe': stdio = ['pipe', 'pipe', 'pipe']; break;
+      case 'inherit': stdio = [0, 1, 2]; break;
+      default: throw new TypeError('Incorrect value of stdio option: ' + stdio);
+    }
+  } else if (!util.isArray(stdio)) {
+    throw new TypeError('Incorrect value of stdio option: ' +
+        util.inspect(stdio));
+  }
+
+  // At least 3 stdio will be created
+  // Don't concat() a new Array() because it would be sparse, and
+  // stdio.reduce() would skip the sparse elements of stdio.
+  // See http://stackoverflow.com/a/5501711/3561
+  while (stdio.length < 3) stdio.push(undefined);
+
+  // Translate stdio into C++-readable form
+  // (i.e. PipeWraps or fds)
+  stdio = stdio.reduce(function(acc, stdio, i) {
+    function cleanup() {
+      acc.filter(function(stdio) {
+        return stdio.type === 'pipe' || stdio.type === 'ipc';
+      }).forEach(function(stdio) {
+        if (stdio.handle)
+          stdio.handle.close();
+      });
+    }
+
+    // Defaults
+    if (util.isNullOrUndefined(stdio)) {
+      stdio = i < 3 ? 'pipe' : 'ignore';
+    }
+
+    if (stdio === null || stdio === 'ignore') {
+      acc.push({type: 'ignore'});
+    } else if (stdio === 'pipe' || util.isNumber(stdio) && stdio < 0) {
+      var a = {
+        type: 'pipe',
+        readable: i === 0,
+        writable: i !== 0
+      };
+
+      if (!sync)
+        a.handle = createPipe();
+
+      acc.push(a);
+    } else if (stdio === 'ipc') {
+      if (sync || !util.isUndefined(ipc)) {
+        // Cleanup previously created pipes
+        cleanup();
+        if (!sync)
+          throw Error('Child process can have only one IPC pipe');
+        else
+          throw Error('You cannot use IPC with synchronous forks');
+      }
+
+      ipc = createPipe(true);
+      ipcFd = i;
+
+      acc.push({
+        type: 'pipe',
+        handle: ipc,
+        ipc: true
+      });
+    } else if (stdio === 'inherit') {
+      acc.push({
+        type: 'inherit',
+        fd: i
+      });
+    } else if (util.isNumber(stdio) || util.isNumber(stdio.fd)) {
+      acc.push({
+        type: 'fd',
+        fd: stdio.fd || stdio
+      });
+    } else if (getHandleWrapType(stdio) || getHandleWrapType(stdio.handle) ||
+               getHandleWrapType(stdio._handle)) {
+      var handle = getHandleWrapType(stdio) ?
+          stdio :
+          getHandleWrapType(stdio.handle) ? stdio.handle : stdio._handle;
+
+      acc.push({
+        type: 'wrap',
+        wrapType: getHandleWrapType(handle),
+        handle: handle
+      });
+    } else if (util.isBuffer(stdio) || util.isString(stdio)) {
+      if (!sync) {
+        cleanup();
+        throw new TypeError('Asynchronous forks do not support Buffer input: ' +
+            util.inspect(stdio));
+      }
+    } else {
+      // Cleanup
+      cleanup();
+      throw new TypeError('Incorrect value for stdio stream: ' +
+          util.inspect(stdio));
+    }
+
+    return acc;
+  }, []);
+
+  return {stdio: stdio, ipc: ipc, ipcFd: ipcFd};
+}
+
+
+function normalizeSpawnArguments(file /*, args, options*/) {
   var args, options;
+
   if (Array.isArray(arguments[1])) {
     args = arguments[1].slice(0);
     options = arguments[2];
+  } else if (arguments[1] !== undefined && !util.isObject(arguments[1])) {
+    throw new TypeError('Incorrect value of args option');
   } else {
     args = [];
     options = arguments[1];
   }
 
+  if (options === undefined)
+    options = {};
+  else if (!util.isObject(options))
+    throw new TypeError('options argument must be an object');
+
   args.unshift(file);
 
-  var env = (options ? options.env : null) || process.env;
+  var env = options.env || process.env;
   var envPairs = [];
+
   for (var key in env) {
     envPairs.push(key + '=' + env[key]);
   }
 
-  var child = new ChildProcess();
-  if (options && options.customFds && !options.stdio) {
-    options.stdio = options.customFds.map(function(fd) {
-      return fd === -1 ? 'pipe' : fd;
-    });
-  }
+  _convertCustomFds(options);
 
-  child.spawn({
+  return {
     file: file,
     args: args,
-    cwd: options ? options.cwd : null,
-    windowsVerbatimArguments: !!(options && options.windowsVerbatimArguments),
-    detached: !!(options && options.detached),
-    envPairs: envPairs,
-    stdio: options ? options.stdio : null,
-    uid: options ? options.uid : null,
-    gid: options ? options.gid : null
+    options: options,
+    envPairs: envPairs
+  };
+}
+
+
+var spawn = exports.spawn = function(/*file, args, options*/) {
+  var opts = normalizeSpawnArguments.apply(null, arguments);
+  var options = opts.options;
+  var child = new ChildProcess();
+
+  child.spawn({
+    file: opts.file,
+    args: opts.args,
+    cwd: options.cwd,
+    windowsVerbatimArguments: !!options.windowsVerbatimArguments,
+    detached: !!options.detached,
+    envPairs: opts.envPairs,
+    stdio: options.stdio,
+    uid: options.uid,
+    gid: options.gid
   });
 
   return child;
@@ -771,6 +1026,7 @@ function ChildProcess() {
   this.signalCode = null;
   this.exitCode = null;
   this.killed = false;
+  this.spawnfile = null;
 
   this._handle = new Process();
   this._handle.owner = this;
@@ -784,9 +1040,10 @@ function ChildProcess() {
     //
     // new in 0.9.x:
     //
-    // - spawn failures are reported with exitCode == -1
+    // - spawn failures are reported with exitCode < 0
     //
-    var err = (exitCode == -1) ? errnoException(process._errno, 'spawn') : null;
+    var syscall = self.spawnfile ? 'spawn ' + self.spawnfile : 'spawn';
+    var err = (exitCode < 0) ? errnoException(exitCode, syscall) : null;
 
     if (signalCode) {
       self.signalCode = signalCode;
@@ -801,7 +1058,10 @@ function ChildProcess() {
     self._handle.close();
     self._handle = null;
 
-    if (exitCode == -1) {
+    if (exitCode < 0) {
+      if (self.spawnfile)
+        err.path = self.spawnfile;
+
       self.emit('error', err);
     } else {
       self.emit('exit', self.exitCode, self.signalCode);
@@ -824,6 +1084,7 @@ util.inherits(ChildProcess, EventEmitter);
 
 
 function flushStdio(subprocess) {
+  if (subprocess.stdio == null) return;
   subprocess.stdio.forEach(function(stream, fd, stdio) {
     if (!stream || !stream.readable || stream._consuming ||
         stream._readableState.flowing)
@@ -851,88 +1112,37 @@ ChildProcess.prototype.spawn = function(options) {
       // If no `stdio` option was given - use default
       stdio = options.stdio || 'pipe';
 
-  // Replace shortcut with an array
-  if (typeof stdio === 'string') {
-    switch (stdio) {
-      case 'ignore': stdio = ['ignore', 'ignore', 'ignore']; break;
-      case 'pipe': stdio = ['pipe', 'pipe', 'pipe']; break;
-      case 'inherit': stdio = [0, 1, 2]; break;
-      default: throw new TypeError('Incorrect value of stdio option: ' + stdio);
-    }
-  } else if (!Array.isArray(stdio)) {
-    throw new TypeError('Incorrect value of stdio option: ' + stdio);
-  }
+  stdio = _validateStdio(stdio, false);
 
-  // At least 3 stdio will be created
-  // Don't concat() a new Array() because it would be sparse, and
-  // stdio.reduce() would skip the sparse elements of stdio.
-  // See http://stackoverflow.com/a/5501711/3561
-  while (stdio.length < 3) stdio.push(undefined);
+  ipc = stdio.ipc;
+  ipcFd = stdio.ipcFd;
+  stdio = options.stdio = stdio.stdio;
 
-  // Translate stdio into C++-readable form
-  // (i.e. PipeWraps or fds)
-  stdio = stdio.reduce(function(acc, stdio, i) {
-    function cleanup() {
-      acc.filter(function(stdio) {
-        return stdio.type === 'pipe' || stdio.type === 'ipc';
-      }).forEach(function(stdio) {
-        stdio.handle.close();
-      });
-    }
-
-    // Defaults
-    if (stdio === undefined || stdio === null) {
-      stdio = i < 3 ? 'pipe' : 'ignore';
-    }
-
-    if (stdio === 'ignore') {
-      acc.push({type: 'ignore'});
-    } else if (stdio === 'pipe' || typeof stdio === 'number' && stdio < 0) {
-      acc.push({type: 'pipe', handle: createPipe()});
-    } else if (stdio === 'ipc') {
-      if (ipc !== undefined) {
-        // Cleanup previously created pipes
-        cleanup();
-        throw Error('Child process can have only one IPC pipe');
-      }
-
-      ipc = createPipe(true);
-      ipcFd = i;
-
-      acc.push({ type: 'pipe', handle: ipc, ipc: true });
-    } else if (typeof stdio === 'number' || typeof stdio.fd === 'number') {
-      acc.push({ type: 'fd', fd: stdio.fd || stdio });
-    } else if (getHandleWrapType(stdio) || getHandleWrapType(stdio.handle) ||
-               getHandleWrapType(stdio._handle)) {
-      var handle = getHandleWrapType(stdio) ?
-          stdio :
-          getHandleWrapType(stdio.handle) ? stdio.handle : stdio._handle;
-
-      acc.push({
-        type: 'wrap',
-        wrapType: getHandleWrapType(handle),
-        handle: handle
-      });
-    } else {
-      // Cleanup
-      cleanup();
-      throw new TypeError('Incorrect value for stdio stream: ' + stdio);
-    }
-
-    return acc;
-  }, []);
-
-  options.stdio = stdio;
-
-  if (ipc !== undefined) {
+  if (!util.isUndefined(ipc)) {
     // Let child process know about opened IPC channel
     options.envPairs = options.envPairs || [];
     options.envPairs.push('NODE_CHANNEL_FD=' + ipcFd);
   }
 
-  var r = this._handle.spawn(options);
+  this.spawnfile = options.file;
 
-  if (r) {
+  var err = this._handle.spawn(options);
+
+  // Run-time errors should emit an error, not throw an exception.
+  if (err === uv.UV_EAGAIN ||
+      err === uv.UV_EMFILE ||
+      err === uv.UV_ENFILE ||
+      err === uv.UV_ENOENT) {
+    process.nextTick(function() {
+      self._handle.onexit(err);
+    });
+    // There is no point in continuing when we've hit EMFILE or ENFILE
+    // because we won't be able to set up the stdio file descriptors.
+    // It's kind of silly that the de facto spec for ENOENT (the test suite)
+    // mandates that stdio _is_ set up, even if there is no process on the
+    // receiving end, but it is what it is.
+    if (err !== uv.UV_ENOENT) return err;
+  } else if (err) {
     // Close all opened fds on error
     stdio.forEach(function(stdio) {
       if (stdio.type === 'pipe') {
@@ -942,7 +1152,7 @@ ChildProcess.prototype.spawn = function(options) {
 
     this._handle.close();
     this._handle = null;
-    throw errnoException(process._errno, 'spawn');
+    throw errnoException(err, 'spawn');
   }
 
   this.pid = this._handle.pid;
@@ -969,39 +1179,24 @@ ChildProcess.prototype.spawn = function(options) {
     }
   });
 
-  this.stdin = stdio.length >= 1 && stdio[0].socket !== undefined ?
+  this.stdin = stdio.length >= 1 && !util.isUndefined(stdio[0].socket) ?
       stdio[0].socket : null;
-  this.stdout = stdio.length >= 2 && stdio[1].socket !== undefined ?
+  this.stdout = stdio.length >= 2 && !util.isUndefined(stdio[1].socket) ?
       stdio[1].socket : null;
-  this.stderr = stdio.length >= 3 && stdio[2].socket !== undefined ?
+  this.stderr = stdio.length >= 3 && !util.isUndefined(stdio[2].socket) ?
       stdio[2].socket : null;
 
   this.stdio = stdio.map(function(stdio) {
-    return stdio.socket === undefined ? null : stdio.socket;
+    return util.isUndefined(stdio.socket) ? null : stdio.socket;
   });
 
   // Add .send() method and start listening for IPC data
-  if (ipc !== undefined) setupChannel(this, ipc);
+  if (!util.isUndefined(ipc)) setupChannel(this, ipc);
 
-  return r;
+  return err;
 };
 
 
-function errnoException(errorno, syscall, errmsg) {
-  // TODO make this more compatible with ErrnoException from src/node.cc
-  // Once all of Node is using this function the ErrnoException from
-  // src/node.cc should be removed.
-  var message = syscall + ' ' + errorno;
-  if (errmsg) {
-    message += ' - ' + errmsg;
-  }
-  var e = new Error(message);
-  e.errno = e.code = errorno;
-  e.syscall = syscall;
-  return e;
-}
-
-
 ChildProcess.prototype.kill = function(sig) {
   var signal;
 
@@ -1017,24 +1212,25 @@ ChildProcess.prototype.kill = function(sig) {
     signal = constants[sig];
   }
 
-  if (signal === undefined) {
+  if (util.isUndefined(signal)) {
     throw new Error('Unknown signal: ' + sig);
   }
 
   if (this._handle) {
-    var r = this._handle.kill(signal);
-    if (r == 0) {
+    var err = this._handle.kill(signal);
+    if (err === 0) {
       /* Success. */
       this.killed = true;
       return true;
-    } else if (process._errno == 'ESRCH') {
+    }
+    if (err === uv.UV_ESRCH) {
       /* Already dead. */
-    } else if (process._errno == 'EINVAL' || process._errno == 'ENOSYS') {
+    } else if (err === uv.UV_EINVAL || err === uv.UV_ENOSYS) {
       /* The underlying platform doesn't support this signal. */
-      throw errnoException(process._errno, 'kill');
+      throw errnoException(err, 'kill');
     } else {
       /* Other error, almost certainly EPERM. */
-      this.emit('error', errnoException(process._errno, 'kill'));
+      this.emit('error', errnoException(err, 'kill'));
     }
   }
 
@@ -1051,3 +1247,140 @@ ChildProcess.prototype.ref = function() {
 ChildProcess.prototype.unref = function() {
   if (this._handle) this._handle.unref();
 };
+
+
+function lookupSignal(signal) {
+  if (typeof signal === 'number')
+    return signal;
+
+  if (!constants)
+    constants = process.binding('constants');
+
+  if (!(signal in constants))
+    throw new Error('Unknown signal: ' + signal);
+
+  return constants[signal];
+}
+
+
+function spawnSync(/*file, args, options*/) {
+  var opts = normalizeSpawnArguments.apply(null, arguments);
+
+  var options = opts.options;
+
+  var i;
+
+  options.file = opts.file;
+  options.args = opts.args;
+  options.envPairs = opts.envPairs;
+
+  if (options.killSignal)
+    options.killSignal = lookupSignal(options.killSignal);
+
+  options.stdio = _validateStdio(options.stdio || 'pipe', true).stdio;
+
+  if (options.input) {
+    var stdin = options.stdio[0] = util._extend({}, options.stdio[0]);
+    stdin.input = options.input;
+  }
+
+  // We may want to pass data in on any given fd, ensure it is a valid buffer
+  for (i = 0; i < options.stdio.length; i++) {
+    var input = options.stdio[i] && options.stdio[i].input;
+    if (input != null) {
+      var pipe = options.stdio[i] = util._extend({}, options.stdio[i]);
+      if (Buffer.isBuffer(input))
+        pipe.input = input;
+      else if (util.isString(input))
+        pipe.input = new Buffer(input, options.encoding);
+      else
+        throw new TypeError(util.format(
+            'stdio[%d] should be Buffer or string not %s',
+            i,
+            typeof input));
+    }
+  }
+
+  if (!spawn_sync)
+    spawn_sync = process.binding('spawn_sync');
+
+  var result = spawn_sync.spawn(options);
+
+  if (result.output && options.encoding) {
+    for (i = 0; i < result.output.length; i++) {
+      if (!result.output[i])
+        continue;
+      result.output[i] = result.output[i].toString(options.encoding);
+    }
+  }
+
+  result.stdout = result.output && result.output[1];
+  result.stderr = result.output && result.output[2];
+
+  if (result.error)
+    result.error = errnoException(result.error, 'spawnSync');
+
+  util._extend(result, opts);
+
+  return result;
+}
+exports.spawnSync = spawnSync;
+
+
+function checkExecSyncError(ret) {
+  if (ret.error || ret.status !== 0) {
+    var err = ret.error;
+    ret.error = null;
+
+    if (!err) {
+      var msg = 'Command failed: ' +
+                (ret.cmd ? ret.cmd : ret.args.join(' ')) +
+                (ret.stderr ? '\n' + ret.stderr.toString() : '');
+      err = new Error(msg);
+    }
+
+    util._extend(err, ret);
+    return err;
+  }
+
+  return false;
+}
+
+
+function execFileSync(/*command, options*/) {
+  var opts = normalizeSpawnArguments.apply(null, arguments);
+  var inheritStderr = !opts.options.stdio;
+
+  var ret = spawnSync(opts.file, opts.args.slice(1), opts.options);
+
+  if (inheritStderr)
+    process.stderr.write(ret.stderr);
+
+  var err = checkExecSyncError(ret);
+
+  if (err)
+    throw err;
+  else
+    return ret.stdout;
+}
+exports.execFileSync = execFileSync;
+
+
+function execSync(/*comand, options*/) {
+  var opts = normalizeExecArgs.apply(null, arguments);
+  var inheritStderr = opts.options ? !opts.options.stdio : true;
+
+  var ret = spawnSync(opts.file, opts.args, opts.options);
+  ret.cmd = opts.cmd;
+
+  if (inheritStderr)
+    process.stderr.write(ret.stderr);
+
+  var err = checkExecSyncError(ret);
+
+  if (err)
+    throw err;
+  else
+    return ret.stdout;
+}
+exports.execSync = execSync;
