@@ -26,129 +26,39 @@ exports.DEFAULT_ENCODING = 'buffer';
 
 try {
   var binding = process.binding('crypto');
-  var SecureContext = binding.SecureContext;
   var randomBytes = binding.randomBytes;
   var pseudoRandomBytes = binding.pseudoRandomBytes;
   var getCiphers = binding.getCiphers;
   var getHashes = binding.getHashes;
-  var crypto = true;
 } catch (e) {
-
-  var crypto = false;
+  throw new Error('node.js not compiled with openssl crypto support.');
 }
 
+var constants = require('constants');
+
 var stream = require('stream');
 var util = require('util');
 
+var DH_GENERATOR = 2;
+
 // This is here because many functions accepted binary strings without
 // any explicit encoding in older versions of node, and we don't want
 // to break them unnecessarily.
 function toBuf(str, encoding) {
   encoding = encoding || 'binary';
-  if (typeof str === 'string') {
+  if (util.isString(str)) {
     if (encoding === 'buffer')
       encoding = 'binary';
     str = new Buffer(str, encoding);
   }
   return str;
 }
+exports._toBuf = toBuf;
 
 
 var assert = require('assert');
 var StringDecoder = require('string_decoder').StringDecoder;
 
-function Credentials(secureProtocol, flags, context) {
-  if (!(this instanceof Credentials)) {
-    return new Credentials(secureProtocol, flags, context);
-  }
-
-  if (!crypto) {
-    throw new Error('node.js not compiled with openssl crypto support.');
-  }
-
-  if (context) {
-    this.context = context;
-  } else {
-    this.context = new SecureContext();
-
-    if (secureProtocol) {
-      this.context.init(secureProtocol);
-    } else {
-      this.context.init();
-    }
-  }
-
-  if (flags) this.context.setOptions(flags);
-}
-
-exports.Credentials = Credentials;
-
-
-exports.createCredentials = function(options, context) {
-  if (!options) options = {};
-
-  var c = new Credentials(options.secureProtocol,
-                          options.secureOptions,
-                          context);
-
-  if (context) return c;
-
-  if (options.key) {
-    if (options.passphrase) {
-      c.context.setKey(options.key, options.passphrase);
-    } else {
-      c.context.setKey(options.key);
-    }
-  }
-
-  if (options.cert) c.context.setCert(options.cert);
-
-  if (options.ciphers) c.context.setCiphers(options.ciphers);
-
-  if (options.ca) {
-    if (Array.isArray(options.ca)) {
-      for (var i = 0, len = options.ca.length; i < len; i++) {
-        c.context.addCACert(options.ca[i]);
-      }
-    } else {
-      c.context.addCACert(options.ca);
-    }
-  } else {
-    c.context.addRootCerts();
-  }
-
-  if (options.crl) {
-    if (Array.isArray(options.crl)) {
-      for (var i = 0, len = options.crl.length; i < len; i++) {
-        c.context.addCRL(options.crl[i]);
-      }
-    } else {
-      c.context.addCRL(options.crl);
-    }
-  }
-
-  if (options.sessionIdContext) {
-    c.context.setSessionIdContext(options.sessionIdContext);
-  }
-
-  if (options.pfx) {
-    var pfx = options.pfx;
-    var passphrase = options.passphrase;
-
-    pfx = toBuf(pfx);
-    if (passphrase)
-      passphrase = toBuf(passphrase);
-
-    if (passphrase) {
-      c.context.loadPKCS12(pfx, passphrase);
-    } else {
-      c.context.loadPKCS12(pfx);
-    }
-  }
-
-  return c;
-};
-
 
 function LazyTransform(options) {
   this._options = options;
@@ -185,35 +95,35 @@ exports.createHash = exports.Hash = Hash;
 function Hash(algorithm, options) {
   if (!(this instanceof Hash))
     return new Hash(algorithm, options);
-  this._binding = new binding.Hash(algorithm);
+  this._handle = new binding.Hash(algorithm);
   LazyTransform.call(this, options);
 }
 
 util.inherits(Hash, LazyTransform);
 
 Hash.prototype._transform = function(chunk, encoding, callback) {
-  this._binding.update(chunk, encoding);
+  this._handle.update(chunk, encoding);
   callback();
 };
 
 Hash.prototype._flush = function(callback) {
   var encoding = this._readableState.encoding || 'buffer';
-  this.push(this._binding.digest(encoding), encoding);
+  this.push(this._handle.digest(encoding), encoding);
   callback();
 };
 
 Hash.prototype.update = function(data, encoding) {
   encoding = encoding || exports.DEFAULT_ENCODING;
-  if (encoding === 'buffer' && typeof data === 'string')
+  if (encoding === 'buffer' && util.isString(data))
     encoding = 'binary';
-  this._binding.update(data, encoding);
+  this._handle.update(data, encoding);
   return this;
 };
 
 
 Hash.prototype.digest = function(outputEncoding) {
   outputEncoding = outputEncoding || exports.DEFAULT_ENCODING;
-  return this._binding.digest(outputEncoding);
+  return this._handle.digest(outputEncoding);
 };
 
 
@@ -222,8 +132,8 @@ exports.createHmac = exports.Hmac = Hmac;
 function Hmac(hmac, key, options) {
   if (!(this instanceof Hmac))
     return new Hmac(hmac, key, options);
-  this._binding = new binding.Hmac();
-  this._binding.init(hmac, toBuf(key));
+  this._handle = new binding.Hmac();
+  this._handle.init(hmac, toBuf(key));
   LazyTransform.call(this, options);
 }
 
@@ -247,9 +157,9 @@ exports.createCipher = exports.Cipher = Cipher;
 function Cipher(cipher, password, options) {
   if (!(this instanceof Cipher))
     return new Cipher(cipher, password, options);
-  this._binding = new binding.Cipher;
+  this._handle = new binding.CipherBase(true);
 
-  this._binding.init(cipher, toBuf(password));
+  this._handle.init(cipher, toBuf(password));
   this._decoder = null;
 
   LazyTransform.call(this, options);
@@ -258,12 +168,17 @@ function Cipher(cipher, password, options) {
 util.inherits(Cipher, LazyTransform);
 
 Cipher.prototype._transform = function(chunk, encoding, callback) {
-  this.push(this._binding.update(chunk, encoding));
+  this.push(this._handle.update(chunk, encoding));
   callback();
 };
 
 Cipher.prototype._flush = function(callback) {
-  this.push(this._binding.final());
+  try {
+    this.push(this._handle.final());
+  } catch (e) {
+    callback(e);
+    return;
+  }
   callback();
 };
 
@@ -271,7 +186,7 @@ Cipher.prototype.update = function(data, inputEncoding, outputEncoding) {
   inputEncoding = inputEncoding || exports.DEFAULT_ENCODING;
   outputEncoding = outputEncoding || exports.DEFAULT_ENCODING;
 
-  var ret = this._binding.update(data, inputEncoding);
+  var ret = this._handle.update(data, inputEncoding);
 
   if (outputEncoding && outputEncoding !== 'buffer') {
     this._decoder = getDecoder(this._decoder, outputEncoding);
@@ -284,7 +199,7 @@ Cipher.prototype.update = function(data, inputEncoding, outputEncoding) {
 
 Cipher.prototype.final = function(outputEncoding) {
   outputEncoding = outputEncoding || exports.DEFAULT_ENCODING;
-  var ret = this._binding.final();
+  var ret = this._handle.final();
 
   if (outputEncoding && outputEncoding !== 'buffer') {
     this._decoder = getDecoder(this._decoder, outputEncoding);
@@ -296,7 +211,7 @@ Cipher.prototype.final = function(outputEncoding) {
 
 
 Cipher.prototype.setAutoPadding = function(ap) {
-  this._binding.setAutoPadding(ap);
+  this._handle.setAutoPadding(ap);
   return this;
 };
 
@@ -306,8 +221,8 @@ exports.createCipheriv = exports.Cipheriv = Cipheriv;
 function Cipheriv(cipher, key, iv, options) {
   if (!(this instanceof Cipheriv))
     return new Cipheriv(cipher, key, iv, options);
-  this._binding = new binding.Cipher();
-  this._binding.initiv(cipher, toBuf(key), toBuf(iv));
+  this._handle = new binding.CipherBase(true);
+  this._handle.initiv(cipher, toBuf(key), toBuf(iv));
   this._decoder = null;
 
   LazyTransform.call(this, options);
@@ -321,6 +236,18 @@ Cipheriv.prototype.update = Cipher.prototype.update;
 Cipheriv.prototype.final = Cipher.prototype.final;
 Cipheriv.prototype.setAutoPadding = Cipher.prototype.setAutoPadding;
 
+Cipheriv.prototype.getAuthTag = function() {
+  return this._handle.getAuthTag();
+};
+
+
+Cipheriv.prototype.setAuthTag = function(tagbuf) {
+  this._handle.setAuthTag(tagbuf);
+};
+
+Cipheriv.prototype.setAAD = function(aadbuf) {
+  this._handle.setAAD(aadbuf);
+};
 
 
 exports.createDecipher = exports.Decipher = Decipher;
@@ -328,8 +255,8 @@ function Decipher(cipher, password, options) {
   if (!(this instanceof Decipher))
     return new Decipher(cipher, password, options);
 
-  this._binding = new binding.Decipher;
-  this._binding.init(cipher, toBuf(password));
+  this._handle = new binding.CipherBase(false);
+  this._handle.init(cipher, toBuf(password));
   this._decoder = null;
 
   LazyTransform.call(this, options);
@@ -351,8 +278,8 @@ function Decipheriv(cipher, key, iv, options) {
   if (!(this instanceof Decipheriv))
     return new Decipheriv(cipher, key, iv, options);
 
-  this._binding = new binding.Decipher;
-  this._binding.initiv(cipher, toBuf(key), toBuf(iv));
+  this._handle = new binding.CipherBase(false);
+  this._handle.initiv(cipher, toBuf(key), toBuf(iv));
   this._decoder = null;
 
   LazyTransform.call(this, options);
@@ -366,6 +293,9 @@ Decipheriv.prototype.update = Cipher.prototype.update;
 Decipheriv.prototype.final = Cipher.prototype.final;
 Decipheriv.prototype.finaltol = Cipher.prototype.final;
 Decipheriv.prototype.setAutoPadding = Cipher.prototype.setAutoPadding;
+Decipheriv.prototype.getAuthTag = Cipheriv.prototype.getAuthTag;
+Decipheriv.prototype.setAuthTag = Cipheriv.prototype.setAuthTag;
+Decipheriv.prototype.setAAD = Cipheriv.prototype.setAAD;
 
 
 
@@ -373,8 +303,8 @@ exports.createSign = exports.Sign = Sign;
 function Sign(algorithm, options) {
   if (!(this instanceof Sign))
     return new Sign(algorithm, options);
-  this._binding = new binding.Sign();
-  this._binding.init(algorithm);
+  this._handle = new binding.Sign();
+  this._handle.init(algorithm);
 
   stream.Writable.call(this, options);
 }
@@ -382,16 +312,21 @@ function Sign(algorithm, options) {
 util.inherits(Sign, stream.Writable);
 
 Sign.prototype._write = function(chunk, encoding, callback) {
-  this._binding.update(chunk, encoding);
+  this._handle.update(chunk, encoding);
   callback();
 };
 
 Sign.prototype.update = Hash.prototype.update;
 
-Sign.prototype.sign = function(key, encoding) {
-  encoding = encoding || exports.DEFAULT_ENCODING;
-  var ret = this._binding.sign(toBuf(key));
+Sign.prototype.sign = function(options, encoding) {
+  if (!options)
+    throw new Error('No key provided to sign');
 
+  var key = options.key || options;
+  var passphrase = options.passphrase || null;
+  var ret = this._handle.sign(toBuf(key), null, passphrase);
+
+  encoding = encoding || exports.DEFAULT_ENCODING;
   if (encoding && encoding !== 'buffer')
     ret = ret.toString(encoding);
 
@@ -405,8 +340,8 @@ function Verify(algorithm, options) {
   if (!(this instanceof Verify))
     return new Verify(algorithm, options);
 
-  this._binding = new binding.Verify;
-  this._binding.init(algorithm);
+  this._handle = new binding.Verify;
+  this._handle.init(algorithm);
 
   stream.Writable.call(this, options);
 }
@@ -418,24 +353,61 @@ Verify.prototype.update = Sign.prototype.update;
 
 Verify.prototype.verify = function(object, signature, sigEncoding) {
   sigEncoding = sigEncoding || exports.DEFAULT_ENCODING;
-  return this._binding.verify(toBuf(object), toBuf(signature, sigEncoding));
+  return this._handle.verify(toBuf(object), toBuf(signature, sigEncoding));
+};
+
+exports.publicEncrypt = function(options, buffer) {
+  var key = options.key || options;
+  var padding = options.padding || constants.RSA_PKCS1_OAEP_PADDING;
+  return binding.publicEncrypt(toBuf(key), buffer, padding);
+};
+
+exports.privateDecrypt = function(options, buffer) {
+  var key = options.key || options;
+  var passphrase = options.passphrase || null;
+  var padding = options.padding || constants.RSA_PKCS1_OAEP_PADDING;
+  return binding.privateDecrypt(toBuf(key), buffer, padding, passphrase);
 };
 
 
 
 exports.createDiffieHellman = exports.DiffieHellman = DiffieHellman;
 
-function DiffieHellman(sizeOrKey, encoding) {
+function DiffieHellman(sizeOrKey, keyEncoding, generator, genEncoding) {
   if (!(this instanceof DiffieHellman))
-    return new DiffieHellman(sizeOrKey, encoding);
+    return new DiffieHellman(sizeOrKey, keyEncoding, generator, genEncoding);
 
-  if (!sizeOrKey)
-    this._binding = new binding.DiffieHellman();
-  else {
-    encoding = encoding || exports.DEFAULT_ENCODING;
-    sizeOrKey = toBuf(sizeOrKey, encoding);
-    this._binding = new binding.DiffieHellman(sizeOrKey);
+  if (!util.isBuffer(sizeOrKey) &&
+      typeof sizeOrKey !== 'number' &&
+      typeof sizeOrKey !== 'string')
+    throw new TypeError('First argument should be number, string or Buffer');
+
+  if (keyEncoding) {
+    if (typeof keyEncoding !== 'string' ||
+        (!Buffer.isEncoding(keyEncoding) && keyEncoding !== 'buffer')) {
+      genEncoding = generator;
+      generator = keyEncoding;
+      keyEncoding = false;
+    }
   }
+
+  keyEncoding = keyEncoding || exports.DEFAULT_ENCODING;
+  genEncoding = genEncoding || exports.DEFAULT_ENCODING;
+
+  if (typeof sizeOrKey !== 'number')
+    sizeOrKey = toBuf(sizeOrKey, keyEncoding);
+
+  if (!generator)
+    generator = DH_GENERATOR;
+  else if (typeof generator !== 'number')
+    generator = toBuf(generator, genEncoding);
+
+  this._handle = new binding.DiffieHellman(sizeOrKey, generator);
+  Object.defineProperty(this, 'verifyError', {
+    enumerable: true,
+    value: this._handle.verifyError,
+    writable: false
+  });
 }
 
 
@@ -446,7 +418,12 @@ exports.DiffieHellmanGroup =
 function DiffieHellmanGroup(name) {
   if (!(this instanceof DiffieHellmanGroup))
     return new DiffieHellmanGroup(name);
-  this._binding = new binding.DiffieHellmanGroup(name);
+  this._handle = new binding.DiffieHellmanGroup(name);
+  Object.defineProperty(this, 'verifyError', {
+    enumerable: true,
+    value: this._handle.verifyError,
+    writable: false
+  });
 }
 
 
@@ -455,7 +432,7 @@ DiffieHellmanGroup.prototype.generateKeys =
     dhGenerateKeys;
 
 function dhGenerateKeys(encoding) {
-  var keys = this._binding.generateKeys();
+  var keys = this._handle.generateKeys();
   encoding = encoding || exports.DEFAULT_ENCODING;
   if (encoding && encoding !== 'buffer')
     keys = keys.toString(encoding);
@@ -470,7 +447,7 @@ DiffieHellmanGroup.prototype.computeSecret =
 function dhComputeSecret(key, inEnc, outEnc) {
   inEnc = inEnc || exports.DEFAULT_ENCODING;
   outEnc = outEnc || exports.DEFAULT_ENCODING;
-  var ret = this._binding.computeSecret(toBuf(key, inEnc));
+  var ret = this._handle.computeSecret(toBuf(key, inEnc));
   if (outEnc && outEnc !== 'buffer')
     ret = ret.toString(outEnc);
   return ret;
@@ -482,7 +459,7 @@ DiffieHellmanGroup.prototype.getPrime =
     dhGetPrime;
 
 function dhGetPrime(encoding) {
-  var prime = this._binding.getPrime();
+  var prime = this._handle.getPrime();
   encoding = encoding || exports.DEFAULT_ENCODING;
   if (encoding && encoding !== 'buffer')
     prime = prime.toString(encoding);
@@ -495,7 +472,7 @@ DiffieHellmanGroup.prototype.getGenerator =
     dhGetGenerator;
 
 function dhGetGenerator(encoding) {
-  var generator = this._binding.getGenerator();
+  var generator = this._handle.getGenerator();
   encoding = encoding || exports.DEFAULT_ENCODING;
   if (encoding && encoding !== 'buffer')
     generator = generator.toString(encoding);
@@ -508,7 +485,7 @@ DiffieHellmanGroup.prototype.getPublicKey =
     dhGetPublicKey;
 
 function dhGetPublicKey(encoding) {
-  var key = this._binding.getPublicKey();
+  var key = this._handle.getPublicKey();
   encoding = encoding || exports.DEFAULT_ENCODING;
   if (encoding && encoding !== 'buffer')
     key = key.toString(encoding);
@@ -521,7 +498,7 @@ DiffieHellmanGroup.prototype.getPrivateKey =
     dhGetPrivateKey;
 
 function dhGetPrivateKey(encoding) {
-  var key = this._binding.getPrivateKey();
+  var key = this._handle.getPrivateKey();
   encoding = encoding || exports.DEFAULT_ENCODING;
   if (encoding && encoding !== 'buffer')
     key = key.toString(encoding);
@@ -531,54 +508,151 @@ function dhGetPrivateKey(encoding) {
 
 DiffieHellman.prototype.setPublicKey = function(key, encoding) {
   encoding = encoding || exports.DEFAULT_ENCODING;
-  this._binding.setPublicKey(toBuf(key, encoding));
+  this._handle.setPublicKey(toBuf(key, encoding));
   return this;
 };
 
 
 DiffieHellman.prototype.setPrivateKey = function(key, encoding) {
   encoding = encoding || exports.DEFAULT_ENCODING;
-  this._binding.setPrivateKey(toBuf(key, encoding));
+  this._handle.setPrivateKey(toBuf(key, encoding));
   return this;
 };
 
 
+function ECDH(curve) {
+  if (!util.isString(curve))
+    throw new TypeError('curve should be a string');
 
-exports.pbkdf2 = function(password, salt, iterations, keylen, callback) {
-  if (typeof callback !== 'function')
+  this._handle = new binding.ECDH(curve);
+}
+
+exports.createECDH = function createECDH(curve) {
+  return new ECDH(curve);
+};
+
+ECDH.prototype.computeSecret = DiffieHellman.prototype.computeSecret;
+ECDH.prototype.setPrivateKey = DiffieHellman.prototype.setPrivateKey;
+ECDH.prototype.setPublicKey = DiffieHellman.prototype.setPublicKey;
+ECDH.prototype.getPrivateKey = DiffieHellman.prototype.getPrivateKey;
+
+ECDH.prototype.generateKeys = function generateKeys(encoding, format) {
+  this._handle.generateKeys();
+
+  return this.getPublicKey(encoding, format);
+};
+
+ECDH.prototype.getPublicKey = function getPublicKey(encoding, format) {
+  var f;
+  if (format) {
+    if (typeof format === 'number')
+      f = format;
+    if (format === 'compressed')
+      f = constants.POINT_CONVERSION_COMPRESSED;
+    else if (format === 'hybrid')
+      f = constants.POINT_CONVERSION_HYBRID;
+    // Default
+    else if (format === 'uncompressed')
+      f = constants.POINT_CONVERSION_UNCOMPRESSED;
+    else
+      throw TypeError('Bad format: ' + format);
+  } else {
+    f = constants.POINT_CONVERSION_UNCOMPRESSED;
+  }
+  var key = this._handle.getPublicKey(f);
+  encoding = encoding || exports.DEFAULT_ENCODING;
+  if (encoding && encoding !== 'buffer')
+    key = key.toString(encoding);
+  return key;
+};
+
+
+
+exports.pbkdf2 = function(password,
+                          salt,
+                          iterations,
+                          keylen,
+                          digest,
+                          callback) {
+  if (util.isFunction(digest)) {
+    callback = digest;
+    digest = undefined;
+  }
+
+  if (!util.isFunction(callback))
     throw new Error('No callback provided to pbkdf2');
 
-  return pbkdf2(password, salt, iterations, keylen, callback);
+  return pbkdf2(password, salt, iterations, keylen, digest, callback);
 };
 
 
-exports.pbkdf2Sync = function(password, salt, iterations, keylen) {
-  return pbkdf2(password, salt, iterations, keylen);
+exports.pbkdf2Sync = function(password, salt, iterations, keylen, digest) {
+  return pbkdf2(password, salt, iterations, keylen, digest);
 };
 
 
-function pbkdf2(password, salt, iterations, keylen, callback) {
+function pbkdf2(password, salt, iterations, keylen, digest, callback) {
   password = toBuf(password);
   salt = toBuf(salt);
 
   if (exports.DEFAULT_ENCODING === 'buffer')
-    return binding.PBKDF2(password, salt, iterations, keylen, callback);
+    return binding.PBKDF2(password, salt, iterations, keylen, digest, callback);
 
   // at this point, we need to handle encodings.
   var encoding = exports.DEFAULT_ENCODING;
   if (callback) {
-    binding.PBKDF2(password, salt, iterations, keylen, function(er, ret) {
+    function next(er, ret) {
       if (ret)
         ret = ret.toString(encoding);
       callback(er, ret);
-    });
+    }
+    binding.PBKDF2(password, salt, iterations, keylen, digest, next);
   } else {
-    var ret = binding.PBKDF2(password, salt, iterations, keylen);
+    var ret = binding.PBKDF2(password, salt, iterations, keylen, digest);
     return ret.toString(encoding);
   }
 }
 
 
+exports.Certificate = Certificate;
+
+function Certificate() {
+  if (!(this instanceof Certificate))
+    return new Certificate();
+
+  this._handle = new binding.Certificate();
+}
+
+
+Certificate.prototype.verifySpkac = function(object) {
+  return this._handle.verifySpkac(object);
+};
+
+
+Certificate.prototype.exportPublicKey = function(object, encoding) {
+  return this._handle.exportPublicKey(toBuf(object, encoding));
+};
+
+
+Certificate.prototype.exportChallenge = function(object, encoding) {
+  return this._handle.exportChallenge(toBuf(object, encoding));
+};
+
+
+exports.setEngine = function setEngine(id, flags) {
+  if (!util.isString(id))
+    throw new TypeError('id should be a string');
+
+  if (flags && !util.isNumber(flags))
+    throw new TypeError('flags should be a number, if present');
+  flags = flags >>> 0;
+
+  // Use provided engine for everything by default
+  if (flags === 0)
+    flags = constants.ENGINE_METHOD_ALL;
+
+  return binding.setEngine(id, flags);
+};
 
 exports.randomBytes = randomBytes;
 exports.pseudoRandomBytes = pseudoRandomBytes;
@@ -603,8 +677,22 @@ function filterDuplicates(names) {
   // for example, 'sha1' instead of 'SHA1'.
   var ctx = {};
   names.forEach(function(name) {
-    if (/^[0-9A-Z\-]+$/.test(name)) name = name.toLowerCase();
-    ctx[name] = true;
+    var key = name;
+    if (/^[0-9A-Z\-]+$/.test(key)) key = key.toLowerCase();
+    if (!ctx.hasOwnProperty(key) || ctx[key] < name)
+      ctx[key] = name;
   });
-  return Object.getOwnPropertyNames(ctx).sort();
+
+  return Object.getOwnPropertyNames(ctx).map(function(key) {
+    return ctx[key];
+  }).sort();
 }
+
+// Legacy API
+exports.__defineGetter__('createCredentials', util.deprecate(function() {
+  return require('tls').createSecureContext;
+}, 'createCredentials() is deprecated, use tls.createSecureContext instead'));
+
+exports.__defineGetter__('Credentials', util.deprecate(function() {
+  return require('tls').SecureContext;
+}, 'Credentials is deprecated, use tls.createSecureContext instead'));
