@@ -25,13 +25,23 @@ var timers = require('timers');
 var util = require('util');
 var assert = require('assert');
 var cares = process.binding('cares_wrap');
+var uv = process.binding('uv');
+var Pipe = process.binding('pipe_wrap').Pipe;
+
+var TCPConnectWrap = process.binding('tcp_wrap').TCPConnectWrap;
+var PipeConnectWrap = process.binding('pipe_wrap').PipeConnectWrap;
+var ShutdownWrap = process.binding('stream_wrap').ShutdownWrap;
+var WriteWrap = process.binding('stream_wrap').WriteWrap;
+
+
+var Buffer = require('buffer').Buffer;
 var cluster;
+var errnoException = util._errnoException;
 
 function noop() {}
 
 // constructor for lazy loading
 function createPipe() {
-  var Pipe = process.binding('pipe_wrap').Pipe;
   return new Pipe();
 }
 
@@ -51,23 +61,10 @@ function createHandle(fd) {
 }
 
 
-var debug;
-if (process.env.NODE_DEBUG && /net/.test(process.env.NODE_DEBUG)) {
-  var pid = process.pid;
-  debug = function(x) {
-    // if console is not set up yet, then skip this.
-    if (!console.error)
-      return;
-    console.error('NET: %d', pid,
-                  util.format.apply(util, arguments).slice(0, 500));
-  };
-} else {
-  debug = function() { };
-}
-
+var debug = util.debuglog('net');
 
 function isPipeName(s) {
-  return typeof s === 'string' && toNumber(s) === false;
+  return util.isString(s) && toNumber(s) === false;
 }
 
 
@@ -90,6 +87,7 @@ exports.createServer = function() {
 //
 exports.connect = exports.createConnection = function() {
   var args = normalizeConnectArgs(arguments);
+  debug('createConnection', args);
   var s = new Socket(args[0]);
   return Socket.prototype.connect.apply(s, args);
 };
@@ -99,7 +97,7 @@ exports.connect = exports.createConnection = function() {
 function normalizeConnectArgs(args) {
   var options = {};
 
-  if (typeof args[0] === 'object') {
+  if (util.isObject(args[0])) {
     // connect(options, [cb])
     options = args[0];
   } else if (isPipeName(args[0])) {
@@ -108,13 +106,13 @@ function normalizeConnectArgs(args) {
   } else {
     // connect(port, [host], [cb])
     options.port = args[0];
-    if (typeof args[1] === 'string') {
+    if (util.isString(args[1])) {
       options.host = args[1];
     }
   }
 
   var cb = args[args.length - 1];
-  return (typeof cb === 'function') ? [options, cb] : [options];
+  return util.isFunction(cb) ? [options, cb] : [options];
 }
 exports._normalizeConnectArgs = normalizeConnectArgs;
 
@@ -129,6 +127,10 @@ function initSocketHandle(self) {
   if (self._handle) {
     self._handle.owner = self;
     self._handle.onread = onread;
+
+    // If handle doesn't support writev - neither do we
+    if (!self._handle.writev)
+      self._writev = null;
   }
 }
 
@@ -136,24 +138,30 @@ function Socket(options) {
   if (!(this instanceof Socket)) return new Socket(options);
 
   this._connecting = false;
+  this._hadError = false;
   this._handle = null;
+  this._host = null;
 
-  switch (typeof options) {
-    case 'number':
-      options = { fd: options }; // Legacy interface.
-      break;
-    case 'undefined':
-      options = {};
-      break;
-  }
+  if (util.isNumber(options))
+    options = { fd: options }; // Legacy interface.
+  else if (util.isUndefined(options))
+    options = {};
 
   stream.Duplex.call(this, options);
 
   if (options.handle) {
     this._handle = options.handle; // private
-  } else if (typeof options.fd !== 'undefined') {
+  } else if (!util.isUndefined(options.fd)) {
     this._handle = createHandle(options.fd);
     this._handle.open(options.fd);
+    if ((options.fd == 1 || options.fd == 2) &&
+        (this._handle instanceof Pipe) &&
+        process.platform === 'win32') {
+      // Make stdout and stderr blocking on Windows
+      var err = this._handle.setBlocking(true);
+      if (err)
+        throw errnoException(err, 'setBlocking');
+    }
     this.readable = options.readable !== false;
     this.writable = options.writable !== false;
   } else {
@@ -161,8 +169,6 @@ function Socket(options) {
     this.readable = this.writable = false;
   }
 
-  this.onend = null;
-
   // shut down the socket when we're finished with it.
   this.on('finish', onSocketFinish);
   this.on('_socketEnd', onSocketEnd);
@@ -180,8 +186,16 @@ function Socket(options) {
 
   // if we have a handle, then start the flow of data into the
   // buffer.  if not, then this will happen when we connect
-  if (this._handle && options.readable !== false)
-    this.read(0);
+  if (this._handle && options.readable !== false) {
+    if (options.pauseOnCreate) {
+      // stop the handle from reading and pause the stream
+      this._handle.reading = false;
+      this._handle.readStop();
+      this._readableState.flowing = false;
+    } else {
+      this.read(0);
+    }
+  }
 }
 util.inherits(Socket, stream.Duplex);
 
@@ -210,12 +224,12 @@ function onSocketFinish() {
   if (!this._handle || !this._handle.shutdown)
     return this.destroy();
 
-  var shutdownReq = this._handle.shutdown();
+  var req = new ShutdownWrap();
+  req.oncomplete = afterShutdown;
+  var err = this._handle.shutdown(req);
 
-  if (!shutdownReq)
-    return this._destroy(errnoException(process._errno, 'shutdown'));
-
-  shutdownReq.oncomplete = afterShutdown;
+  if (err)
+    return this._destroy(errnoException(err, 'shutdown'));
 }
 
 
@@ -267,7 +281,7 @@ function onSocketEnd() {
 // of the other side sending a FIN.  The standard 'write after end'
 // is overly vague, and makes it seem like the user's code is to blame.
 function writeAfterFIN(chunk, encoding, cb) {
-  if (typeof encoding === 'function') {
+  if (util.isFunction(encoding)) {
     cb = encoding;
     encoding = null;
   }
@@ -277,7 +291,7 @@ function writeAfterFIN(chunk, encoding, cb) {
   var self = this;
   // TODO: defer error events consistently everywhere, not just the cb
   self.emit('error', er);
-  if (typeof cb === 'function') {
+  if (util.isFunction(cb)) {
     process.nextTick(function() {
       cb(er);
     });
@@ -306,17 +320,17 @@ Socket.prototype.listen = function() {
 
 
 Socket.prototype.setTimeout = function(msecs, callback) {
-  if (msecs > 0 && isFinite(msecs)) {
+  if (msecs === 0) {
+    timers.unenroll(this);
+    if (callback) {
+      this.removeListener('timeout', callback);
+    }
+  } else {
     timers.enroll(this, msecs);
     timers._unrefActive(this);
     if (callback) {
       this.once('timeout', callback);
     }
-  } else if (msecs === 0) {
-    timers.unenroll(this);
-    if (callback) {
-      this.removeListener('timeout', callback);
-    }
   }
 };
 
@@ -330,7 +344,7 @@ Socket.prototype._onTimeout = function() {
 Socket.prototype.setNoDelay = function(enable) {
   // backwards compatibility: assume true when `enable` is omitted
   if (this._handle && this._handle.setNoDelay)
-    this._handle.setNoDelay(typeof enable === 'undefined' ? true : !!enable);
+    this._handle.setNoDelay(util.isUndefined(enable) ? true : !!enable);
 };
 
 
@@ -341,10 +355,7 @@ Socket.prototype.setKeepAlive = function(setting, msecs) {
 
 
 Socket.prototype.address = function() {
-  if (this._handle && this._handle.getsockname) {
-    return this._handle.getsockname();
-  }
-  return null;
+  return this._getsockname();
 };
 
 
@@ -385,9 +396,9 @@ Socket.prototype._read = function(n) {
     // not already reading, start the flow
     debug('Socket._read readStart');
     this._handle.reading = true;
-    var r = this._handle.readStart();
-    if (r)
-      this._destroy(errnoException(process._errno, 'read'));
+    var err = this._handle.readStart();
+    if (err)
+      this._destroy(errnoException(err, 'read'));
   }
 };
 
@@ -493,17 +504,16 @@ Socket.prototype.destroy = function(exception) {
 
 // This function is called whenever the handle gets a
 // buffer, or when there's an error reading.
-function onread(buffer, offset, length) {
+function onread(nread, buffer) {
   var handle = this;
   var self = handle.owner;
   assert(handle === self._handle, 'handle != self._handle');
 
   timers._unrefActive(self);
 
-  var end = offset + length;
-  debug('onread', process._errno, offset, length, end);
+  debug('onread', nread);
 
-  if (buffer) {
+  if (nread > 0) {
     debug('got data');
 
     // read success.
@@ -511,52 +521,49 @@ function onread(buffer, offset, length) {
     // will prevent this from being called again until _read() gets
     // called again.
 
-    // if we didn't get any bytes, that doesn't necessarily mean EOF.
-    // wait for the next one.
-    if (offset === end) {
-      debug('not any data, keep waiting');
-      return;
-    }
-
     // if it's not enough data, we'll just call handle.readStart()
     // again right away.
-    self.bytesRead += length;
+    self.bytesRead += nread;
 
     // Optimization: emit the original buffer with end points
-    var ret = true;
-    if (self.ondata) self.ondata(buffer, offset, end);
-    else ret = self.push(buffer.slice(offset, end));
+    var ret = self.push(buffer);
 
     if (handle.reading && !ret) {
       handle.reading = false;
       debug('readStop');
-      var r = handle.readStop();
-      if (r)
-        self._destroy(errnoException(process._errno, 'read'));
+      var err = handle.readStop();
+      if (err)
+        self._destroy(errnoException(err, 'read'));
     }
-
-  } else if (process._errno == 'EOF') {
-    debug('EOF');
-
-    if (self._readableState.length === 0) {
-      self.readable = false;
-      maybeDestroy(self);
-    }
-
-    if (self.onend) self.once('end', self.onend);
-
-    // push a null to signal the end of data.
-    self.push(null);
-
-    // internal end event so that we know that the actual socket
-    // is no longer readable, and we can start the shutdown
-    // procedure. No need to wait for all the data to be consumed.
-    self.emit('_socketEnd');
-  } else {
-    debug('error', process._errno);
-    // Error
-    self._destroy(errnoException(process._errno, 'read'));
+    return;
   }
+
+  // if we didn't get any bytes, that doesn't necessarily mean EOF.
+  // wait for the next one.
+  if (nread === 0) {
+    debug('not any data, keep waiting');
+    return;
+  }
+
+  // Error, possibly EOF.
+  if (nread !== uv.UV_EOF) {
+    return self._destroy(errnoException(nread, 'read'));
+  }
+
+  debug('EOF');
+
+  if (self._readableState.length === 0) {
+    self.readable = false;
+    maybeDestroy(self);
+  }
+
+  // push a null to signal the end of data.
+  self.push(null);
+
+  // internal end event so that we know that the actual socket
+  // is no longer readable, and we can start the shutdown
+  // procedure. No need to wait for all the data to be consumed.
+  self.emit('_socketEnd');
 }
 
 
@@ -565,11 +572,10 @@ Socket.prototype._getpeername = function() {
     return {};
   }
   if (!this._peername) {
-    this._peername = this._handle.getpeername();
-    // getpeername() returns null on error
-    if (this._peername === null) {
-      return {};
-    }
+    var out = {};
+    var err = this._handle.getpeername(out);
+    if (err) return {};  // FIXME(bnoordhuis) Throw?
+    this._peername = out;
   }
   return this._peername;
 };
@@ -579,6 +585,9 @@ Socket.prototype.__defineGetter__('remoteAddress', function() {
   return this._getpeername().address;
 });
 
+Socket.prototype.__defineGetter__('remoteFamily', function() {
+  return this._getpeername().family;
+});
 
 Socket.prototype.__defineGetter__('remotePort', function() {
   return this._getpeername().port;
@@ -590,10 +599,10 @@ Socket.prototype._getsockname = function() {
     return {};
   }
   if (!this._sockname) {
-    this._sockname = this._handle.getsockname();
-    if (this._sockname === null) {
-      return {};
-    }
+    var out = {};
+    var err = this._handle.getsockname(out);
+    if (err) return {};  // FIXME(bnoordhuis) Throw?
+    this._sockname = out;
   }
   return this._sockname;
 };
@@ -610,13 +619,14 @@ Socket.prototype.__defineGetter__('localPort', function() {
 
 
 Socket.prototype.write = function(chunk, encoding, cb) {
-  if (typeof chunk !== 'string' && !Buffer.isBuffer(chunk))
+  'use strict';
+  if (!util.isString(chunk) && !util.isBuffer(chunk))
     throw new TypeError('invalid data');
   return stream.Duplex.prototype.write.apply(this, arguments);
 };
 
 
-Socket.prototype._write = function(data, encoding, cb) {
+Socket.prototype._writeGeneric = function(writev, data, encoding, cb) {
   // If we are still connecting, then buffer this for later.
   // The Writable logic will buffer up any more writes while
   // waiting for this one to be done.
@@ -624,7 +634,7 @@ Socket.prototype._write = function(data, encoding, cb) {
     this._pendingData = data;
     this._pendingEncoding = encoding;
     this.once('connect', function() {
-      this._write(data, encoding, cb);
+      this._writeGeneric(writev, data, encoding, cb);
     });
     return;
   }
@@ -638,43 +648,81 @@ Socket.prototype._write = function(data, encoding, cb) {
     return false;
   }
 
-  var enc = Buffer.isBuffer(data) ? 'buffer' : encoding;
-  var writeReq = createWriteReq(this._handle, data, enc);
+  var req = new WriteWrap();
+  req.oncomplete = afterWrite;
+  req.async = false;
+  var err;
 
-  if (!writeReq || typeof writeReq !== 'object')
-    return this._destroy(errnoException(process._errno, 'write'), cb);
+  if (writev) {
+    var chunks = new Array(data.length << 1);
+    for (var i = 0; i < data.length; i++) {
+      var entry = data[i];
+      var chunk = entry.chunk;
+      var enc = entry.encoding;
+      chunks[i * 2] = chunk;
+      chunks[i * 2 + 1] = enc;
+    }
+    err = this._handle.writev(req, chunks);
 
-  writeReq.oncomplete = afterWrite;
-  this._bytesDispatched += writeReq.bytes;
+    // Retain chunks
+    if (err === 0) req._chunks = chunks;
+  } else {
+    var enc;
+    if (util.isBuffer(data)) {
+      req.buffer = data;  // Keep reference alive.
+      enc = 'buffer';
+    } else {
+      enc = encoding;
+    }
+    err = createWriteReq(req, this._handle, data, enc);
+  }
+
+  if (err)
+    return this._destroy(errnoException(err, 'write', req.error), cb);
+
+  this._bytesDispatched += req.bytes;
 
   // If it was entirely flushed, we can write some more right now.
   // However, if more is left in the queue, then wait until that clears.
-  if (this._handle.writeQueueSize === 0)
-    cb();
+  if (req.async && this._handle.writeQueueSize != 0)
+    req.cb = cb;
   else
-    writeReq.cb = cb;
+    cb();
 };
 
-function createWriteReq(handle, data, encoding) {
+
+Socket.prototype._writev = function(chunks, cb) {
+  this._writeGeneric(true, chunks, '', cb);
+};
+
+
+Socket.prototype._write = function(data, encoding, cb) {
+  this._writeGeneric(false, data, encoding, cb);
+};
+
+function createWriteReq(req, handle, data, encoding) {
   switch (encoding) {
+    case 'binary':
+      return handle.writeBinaryString(req, data);
+
     case 'buffer':
-      return handle.writeBuffer(data);
+      return handle.writeBuffer(req, data);
 
     case 'utf8':
     case 'utf-8':
-      return handle.writeUtf8String(data);
+      return handle.writeUtf8String(req, data);
 
     case 'ascii':
-      return handle.writeAsciiString(data);
+      return handle.writeAsciiString(req, data);
 
     case 'ucs2':
     case 'ucs-2':
     case 'utf16le':
     case 'utf-16le':
-      return handle.writeUcs2String(data);
+      return handle.writeUcs2String(req, data);
 
     default:
-      return handle.writeBuffer(new Buffer(data, encoding));
+      return handle.writeBuffer(req, new Buffer(data, encoding));
   }
 }
 
@@ -685,15 +733,15 @@ Socket.prototype.__defineGetter__('bytesWritten', function() {
       data = this._pendingData,
       encoding = this._pendingEncoding;
 
-  state.buffer.forEach(function(el) {
-    if (Buffer.isBuffer(el.chunk))
+  state.getBuffer().forEach(function(el) {
+    if (util.isBuffer(el.chunk))
       bytes += el.chunk.length;
     else
       bytes += Buffer.byteLength(el.chunk, el.encoding);
   });
 
   if (data) {
-    if (Buffer.isBuffer(data))
+    if (util.isBuffer(data))
       bytes += data.length;
     else
       bytes += Buffer.byteLength(data, encoding);
@@ -703,11 +751,10 @@ Socket.prototype.__defineGetter__('bytesWritten', function() {
 });
 
 
-function afterWrite(status, handle, req) {
+function afterWrite(status, handle, req, err) {
   var self = handle.owner;
-  var state = self._writableState;
   if (self !== process.stderr && self !== process.stdout)
-    debug('afterWrite', status, req);
+    debug('afterWrite', status);
 
   // callback may come after call to destroy.
   if (self.destroyed) {
@@ -715,9 +762,10 @@ function afterWrite(status, handle, req) {
     return;
   }
 
-  if (status) {
-    debug('write failure', errnoException(process._errno, 'write'));
-    self._destroy(errnoException(process._errno, 'write'), req.cb);
+  if (status < 0) {
+    var ex = errnoException(status, 'write', err);
+    debug('write failure', ex);
+    self._destroy(ex, req.cb);
     return;
   }
 
@@ -731,39 +779,58 @@ function afterWrite(status, handle, req) {
 }
 
 
-function connect(self, address, port, addressType, localAddress) {
+function connect(self, address, port, addressType, localAddress, localPort) {
   // TODO return promise from Socket.prototype.connect which
   // wraps _connectReq.
 
   assert.ok(self._connecting);
 
-  if (localAddress) {
-    var r;
-    if (addressType == 6) {
-      r = self._handle.bind6(localAddress);
+  var err;
+
+  if (localAddress || localPort) {
+    var bind;
+
+    if (addressType === 4) {
+      localAddress = localAddress || '0.0.0.0';
+      bind = self._handle.bind;
+    } else if (addressType === 6) {
+      localAddress = localAddress || '::';
+      bind = self._handle.bind6;
     } else {
-      r = self._handle.bind(localAddress);
+      self._destroy(new TypeError('Invalid addressType: ' + addressType));
+      return;
     }
 
-    if (r) {
-      self._destroy(errnoException(process._errno, 'bind'));
+    debug('binding to localAddress: %s and localPort: %d',
+          localAddress,
+          localPort);
+
+    bind = bind.bind(self._handle);
+    err = bind(localAddress, localPort);
+
+    if (err) {
+      self._destroy(errnoException(err, 'bind'));
       return;
     }
   }
 
-  var connectReq;
-  if (addressType == 6) {
-    connectReq = self._handle.connect6(address, port);
-  } else if (addressType == 4) {
-    connectReq = self._handle.connect(address, port);
+  if (addressType === 6 || addressType === 4) {
+    var req = new TCPConnectWrap();
+    req.oncomplete = afterConnect;
+
+    if (addressType === 4)
+      err = self._handle.connect(req, address, port);
+    else
+      err = self._handle.connect6(req, address, port);
+
   } else {
-    connectReq = self._handle.connect(address, afterConnect);
+    var req = new PipeConnectWrap();
+    req.oncomplete = afterConnect;
+    err = self._handle.connect(req, address, afterConnect);
   }
 
-  if (connectReq !== null) {
-    connectReq.oncomplete = afterConnect;
-  } else {
-    self._destroy(errnoException(process._errno, 'connect'));
+  if (err) {
+    self._destroy(errnoException(err, 'connect'));
   }
 }
 
@@ -772,7 +839,7 @@ Socket.prototype.connect = function(options, cb) {
   if (this.write !== Socket.prototype.write)
     this.write = Socket.prototype.write;
 
-  if (typeof options !== 'object') {
+  if (!util.isObject(options)) {
     // Old API:
     // connect(port, [host], [cb])
     // connect(path, [cb]);
@@ -787,19 +854,21 @@ Socket.prototype.connect = function(options, cb) {
     this._writableState.ended = false;
     this._writableState.ending = false;
     this._writableState.finished = false;
+    this._writableState.errorEmitted = false;
     this.destroyed = false;
     this._handle = null;
   }
 
   var self = this;
   var pipe = !!options.path;
+  debug('pipe', pipe, options.path);
 
   if (!this._handle) {
     this._handle = pipe ? createPipe() : createTCP();
     initSocketHandle(this);
   }
 
-  if (typeof cb === 'function') {
+  if (util.isFunction(cb)) {
     self.once('connect', cb);
   }
 
@@ -811,14 +880,35 @@ Socket.prototype.connect = function(options, cb) {
   if (pipe) {
     connect(self, options.path);
 
-  } else if (!options.host) {
-    debug('connect: missing host');
-    connect(self, '127.0.0.1', options.port, 4);
-
   } else {
-    var host = options.host;
+    var dns = require('dns');
+    var host = options.host || 'localhost';
+    var port = options.port | 0;
+    var localAddress = options.localAddress;
+    var localPort = options.localPort;
+    var dnsopts = {
+      family: options.family,
+      hints: 0
+    };
+
+    if (localAddress && !exports.isIP(localAddress))
+      throw new TypeError('localAddress must be a valid IP: ' + localAddress);
+
+    if (localPort && !util.isNumber(localPort))
+      throw new TypeError('localPort should be a number: ' + localPort);
+
+    if (port <= 0 || port > 65535)
+      throw new RangeError('port should be > 0 and < 65536: ' + port);
+
+    if (dnsopts.family !== 4 && dnsopts.family !== 6)
+      dnsopts.hints = dns.ADDRCONFIG | dns.V4MAPPED;
+
     debug('connect: find host ' + host);
-    require('dns').lookup(host, function(err, ip, addressType) {
+    debug('connect: dns options ' + dnsopts);
+    self._host = host;
+    dns.lookup(host, dnsopts, function(err, ip, addressType) {
+      self.emit('lookup', err, ip, addressType);
+
       // It's possible we were destroyed while looking this up.
       // XXX it would be great if we could cancel the promise returned by
       // the look up.
@@ -835,14 +925,12 @@ Socket.prototype.connect = function(options, cb) {
         });
       } else {
         timers._unrefActive(self);
-
-        addressType = addressType || 4;
-
-        // node_net.cc handles null host names graciously but user land
-        // expects remoteAddress to have a meaningful value
-        ip = ip || (addressType === 4 ? '127.0.0.1' : '0:0:0:0:0:0:0:1');
-
-        connect(self, ip, options.port, addressType, options.localAddress);
+        connect(self,
+                ip,
+                port,
+                addressType,
+                localAddress,
+                localPort);
       }
     });
   }
@@ -887,29 +975,16 @@ function afterConnect(status, handle, req, readable, writable) {
 
     // start the first read, or get an immediate EOF.
     // this doesn't actually consume any bytes, because len=0.
-    if (readable)
+    if (readable && !self.isPaused())
       self.read(0);
 
   } else {
     self._connecting = false;
-    self._destroy(errnoException(process._errno, 'connect'));
+    self._destroy(errnoException(status, 'connect'));
   }
 }
 
 
-function errnoException(errorno, syscall) {
-  // TODO make this more compatible with ErrnoException from src/node.cc
-  // Once all of Node is using this function the ErrnoException from
-  // src/node.cc should be removed.
-  var e = new Error(syscall + ' ' + errorno);
-  e.errno = e.code = errorno;
-  e.syscall = syscall;
-  return e;
-}
-
-
-
-
 function Server(/* [ options, ] listener */) {
   if (!(this instanceof Server)) return new Server(arguments[0], arguments[1]);
   events.EventEmitter.call(this);
@@ -918,13 +993,13 @@ function Server(/* [ options, ] listener */) {
 
   var options;
 
-  if (typeof arguments[0] == 'function') {
+  if (util.isFunction(arguments[0])) {
     options = {};
     self.on('connection', arguments[0]);
   } else {
     options = arguments[0] || {};
 
-    if (typeof arguments[1] == 'function') {
+    if (util.isFunction(arguments[1])) {
       self.on('connection', arguments[1]);
     }
   }
@@ -942,7 +1017,7 @@ function Server(/* [ options, ] listener */) {
     set: util.deprecate(function(val) {
       return (self._connections = val);
     }, 'connections property is deprecated. Use getConnections() method'),
-    configurable: true, enumerable: true
+    configurable: true, enumerable: false
   });
 
   this._handle = null;
@@ -950,6 +1025,7 @@ function Server(/* [ options, ] listener */) {
   this._slaves = [];
 
   this.allowHalfOpen = options.allowHalfOpen || false;
+  this.pauseOnConnect = !!options.pauseOnConnect;
 }
 util.inherits(Server, events.EventEmitter);
 exports.Server = Server;
@@ -957,29 +1033,34 @@ exports.Server = Server;
 
 function toNumber(x) { return (x = Number(x)) >= 0 ? x : false; }
 
+function _listen(handle, backlog) {
+  // Use a backlog of 512 entries. We pass 511 to the listen() call because
+  // the kernel does: backlogsize = roundup_pow_of_two(backlogsize + 1);
+  // which will thus give us a backlog of 512 entries.
+  return handle.listen(backlog || 511);
+}
 
 var createServerHandle = exports._createServerHandle =
     function(address, port, addressType, fd) {
-  var r = 0;
+  var err = 0;
   // assign handle in listen, and clean up if bind or listen fails
   var handle;
 
-  if (typeof fd === 'number' && fd >= 0) {
+  var isTCP = false;
+  if (util.isNumber(fd) && fd >= 0) {
     try {
       handle = createHandle(fd);
     }
     catch (e) {
       // Not a fd we can listen on.  This will trigger an error.
       debug('listen invalid fd=' + fd + ': ' + e.message);
-      process._errno = 'EINVAL'; // hack, callers expect that errno is set
-      return null;
+      return uv.UV_EINVAL;
     }
     handle.open(fd);
     handle.readable = true;
     handle.writable = true;
-    return handle;
-
-  } else if (port == -1 && addressType == -1) {
+    assert(!address && !port);
+  } else if (port === -1 && addressType === -1) {
     handle = createPipe();
     if (process.platform === 'win32') {
       var instances = parseInt(process.env.NODE_PENDING_PIPE_INSTANCES);
@@ -989,20 +1070,29 @@ var createServerHandle = exports._createServerHandle =
     }
   } else {
     handle = createTCP();
+    isTCP = true;
   }
 
-  if (address || port) {
-    debug('bind to ' + address);
-    if (addressType == 6) {
-      r = handle.bind6(address, port);
+  if (address || port || isTCP) {
+    debug('bind to ' + (address || 'anycast'));
+    if (!address) {
+      // Try binding to ipv6 first
+      err = handle.bind6('::', port);
+      if (err) {
+        handle.close();
+        // Fallback to ipv4
+        return createServerHandle('0.0.0.0', port);
+      }
+    } else if (addressType === 6) {
+      err = handle.bind6(address, port);
     } else {
-      r = handle.bind(address, port);
+      err = handle.bind(address, port);
     }
   }
 
-  if (r) {
+  if (err) {
     handle.close();
-    handle = null;
+    return err;
   }
 
   return handle;
@@ -1012,20 +1102,20 @@ var createServerHandle = exports._createServerHandle =
 Server.prototype._listen2 = function(address, port, addressType, backlog, fd) {
   debug('listen2', address, port, addressType, backlog);
   var self = this;
-  var r = 0;
 
   // If there is not yet a handle, we need to create one and bind.
   // In the case of a server sent via IPC, we don't need to do this.
   if (!self._handle) {
     debug('_listen2: create a handle');
-    self._handle = createServerHandle(address, port, addressType, fd);
-    if (!self._handle) {
-      var error = errnoException(process._errno, 'listen');
+    var rval = createServerHandle(address, port, addressType, fd);
+    if (util.isNumber(rval)) {
+      var error = errnoException(rval, 'listen');
       process.nextTick(function() {
         self.emit('error', error);
       });
       return;
     }
+    self._handle = rval;
   } else {
     debug('_listen2: have a handle already');
   }
@@ -1033,13 +1123,10 @@ Server.prototype._listen2 = function(address, port, addressType, backlog, fd) {
   self._handle.onconnection = onconnection;
   self._handle.owner = self;
 
-  // Use a backlog of 512 entries. We pass 511 to the listen() call because
-  // the kernel does: backlogsize = roundup_pow_of_two(backlogsize + 1);
-  // which will thus give us a backlog of 512 entries.
-  r = self._handle.listen(backlog || 511);
+  var err = _listen(self._handle, backlog);
 
-  if (r) {
-    var ex = errnoException(process._errno, 'listen');
+  if (err) {
+    var ex = errnoException(err, 'listen');
     self._handle.close();
     self._handle = null;
     process.nextTick(function() {
@@ -1052,42 +1139,47 @@ Server.prototype._listen2 = function(address, port, addressType, backlog, fd) {
   this._connectionKey = addressType + ':' + address + ':' + port;
 
   process.nextTick(function() {
-    self.emit('listening');
+    // ensure handle hasn't closed
+    if (self._handle)
+      self.emit('listening');
   });
 };
 
 
-function listen(self, address, port, addressType, backlog, fd) {
+function listen(self, address, port, addressType, backlog, fd, exclusive) {
+  exclusive = !!exclusive;
+
   if (!cluster) cluster = require('cluster');
 
-  if (cluster.isMaster) {
+  if (cluster.isMaster || exclusive) {
     self._listen2(address, port, addressType, backlog, fd);
     return;
   }
 
-  cluster._getServer(self, address, port, addressType, fd, function(handle,
-                                                                    err) {
-        // EACCESS and friends
-        if (err) {
-          self.emit('error', errnoException(err, 'bind'));
-          return;
-        }
+  cluster._getServer(self, address, port, addressType, fd, cb);
 
-        // Some operating systems (notably OS X and Solaris) don't report
-        // EADDRINUSE errors right away. libuv mimics that behavior for the
-        // sake of platform consistency but that means we have have a socket on
-        // our hands that is not actually bound. That's why we check if the
-        // actual port matches what we requested and if not, raise an error.
-        // The exception is when port == 0 because that means "any random
-        // port".
-        if (port && handle.getsockname && port != handle.getsockname().port) {
-          self.emit('error', errnoException('EADDRINUSE', 'bind'));
-          return;
-        }
+  function cb(err, handle) {
+    // EADDRINUSE may not be reported until we call listen(). To complicate
+    // matters, a failed bind() followed by listen() will implicitly bind to
+    // a random port. Ergo, check that the socket is bound to the expected
+    // port before calling listen().
+    //
+    // FIXME(bnoordhuis) Doesn't work for pipe handles, they don't have a
+    // getsockname() method. Non-issue for now, the cluster module doesn't
+    // really support pipes anyway.
+    if (err === 0 && port > 0 && handle.getsockname) {
+      var out = {};
+      err = handle.getsockname(out);
+      if (err === 0 && port !== out.port)
+        err = uv.UV_EADDRINUSE;
+    }
 
-        self._handle = handle;
-        self._listen2(address, port, addressType, backlog, fd);
-      });
+    if (err)
+      return self.emit('error', errnoException(err, 'bind'));
+
+    self._handle = handle;
+    self._listen2(address, port, addressType, backlog, fd);
+  }
 }
 
 
@@ -1095,7 +1187,7 @@ Server.prototype.listen = function() {
   var self = this;
 
   var lastArg = arguments[arguments.length - 1];
-  if (typeof lastArg == 'function') {
+  if (util.isFunction(lastArg)) {
     self.once('listening', lastArg);
   }
 
@@ -1107,52 +1199,71 @@ Server.prototype.listen = function() {
 
   var TCP = process.binding('tcp_wrap').TCP;
 
-  if (arguments.length == 0 || typeof arguments[0] == 'function') {
+  if (arguments.length === 0 || util.isFunction(arguments[0])) {
     // Bind to a random port.
-    listen(self, '0.0.0.0', 0, null, backlog);
-
-  } else if (arguments[0] && typeof arguments[0] === 'object') {
+    listen(self, null, 0, null, backlog);
+  } else if (util.isObject(arguments[0])) {
     var h = arguments[0];
-    if (h._handle) {
-      h = h._handle;
-    } else if (h.handle) {
-      h = h.handle;
-    }
+    h = h._handle || h.handle || h;
+
     if (h instanceof TCP) {
       self._handle = h;
       listen(self, null, -1, -1, backlog);
-    } else if (typeof h.fd === 'number' && h.fd >= 0) {
+    } else if (util.isNumber(h.fd) && h.fd >= 0) {
       listen(self, null, null, null, backlog, h.fd);
     } else {
-      throw new Error('Invalid listen argument: ' + h);
+      // The first argument is a configuration object
+      if (h.backlog)
+        backlog = h.backlog;
+
+      if (util.isNumber(h.port)) {
+        if (h.host)
+          listenAfterLookup(h.port, h.host, backlog, h.exclusive);
+        else
+          listen(self, null, h.port, 4, backlog, undefined, h.exclusive);
+      } else if (h.path && isPipeName(h.path)) {
+        var pipeName = self._pipeName = h.path;
+        listen(self, pipeName, -1, -1, backlog, undefined, h.exclusive);
+      } else {
+        throw new Error('Invalid listen argument: ' + h);
+      }
     }
   } else if (isPipeName(arguments[0])) {
     // UNIX socket or Windows pipe.
     var pipeName = self._pipeName = arguments[0];
     listen(self, pipeName, -1, -1, backlog);
 
-  } else if (typeof arguments[1] == 'undefined' ||
-             typeof arguments[1] == 'function' ||
-             typeof arguments[1] == 'number') {
+  } else if (util.isUndefined(arguments[1]) ||
+             util.isFunction(arguments[1]) ||
+             util.isNumber(arguments[1])) {
     // The first argument is the port, no IP given.
-    listen(self, '0.0.0.0', port, 4, backlog);
+    listen(self, null, port, 4, backlog);
 
   } else {
     // The first argument is the port, the second an IP.
-    require('dns').lookup(arguments[1], function(err, ip, addressType) {
+    listenAfterLookup(port, arguments[1], backlog);
+  }
+
+  function listenAfterLookup(port, address, backlog, exclusive) {
+    require('dns').lookup(address, function(err, ip, addressType) {
       if (err) {
         self.emit('error', err);
       } else {
-        listen(self, ip || '0.0.0.0', port, ip ? addressType : 4, backlog);
+        addressType = ip ? addressType : 4;
+        listen(self, ip, port, addressType, backlog, undefined, exclusive);
       }
     });
   }
+
   return self;
 };
 
 Server.prototype.address = function() {
   if (this._handle && this._handle.getsockname) {
-    return this._handle.getsockname();
+    var out = {};
+    var err = this._handle.getsockname(out);
+    // TODO(bnoordhuis) Check err and throw?
+    return out;
   } else if (this._pipeName) {
     return this._pipeName;
   } else {
@@ -1160,14 +1271,14 @@ Server.prototype.address = function() {
   }
 };
 
-function onconnection(clientHandle) {
+function onconnection(err, clientHandle) {
   var handle = this;
   var self = handle.owner;
 
   debug('onconnection');
 
-  if (!clientHandle) {
-    self.emit('error', errnoException(process._errno, 'accept'));
+  if (err) {
+    self.emit('error', errnoException(err, 'accept'));
     return;
   }
 
@@ -1178,7 +1289,8 @@ function onconnection(clientHandle) {
 
   var socket = new Socket({
     handle: clientHandle,
-    allowHalfOpen: self.allowHalfOpen
+    allowHalfOpen: self.allowHalfOpen,
+    pauseOnCreate: self.pauseOnConnect
   });
   socket.readable = socket.writable = true;
 
@@ -1231,16 +1343,20 @@ Server.prototype.close = function(cb) {
     self._emitCloseIfDrained();
   }
 
-  if (!this._handle) {
-    // Throw error. Follows net_legacy behaviour.
-    throw new Error('Not running');
+  if (cb) {
+    if (!this._handle) {
+      this.once('close', function() {
+        cb(new Error('Not running'));
+      });
+    } else {
+      this.once('close', cb);
+    }
   }
 
-  if (cb) {
-    this.once('close', cb);
+  if (this._handle) {
+    this._handle.close();
+    this._handle = null;
   }
-  this._handle.close();
-  this._handle = null;
 
   if (this._usingSlaves) {
     var self = this,
@@ -1317,11 +1433,11 @@ if (process.platform === 'win32') {
   var simultaneousAccepts;
 
   exports._setSimultaneousAccepts = function(handle) {
-    if (typeof handle === 'undefined') {
+    if (util.isUndefined(handle)) {
       return;
     }
 
-    if (typeof simultaneousAccepts === 'undefined') {
+    if (util.isUndefined(simultaneousAccepts)) {
       simultaneousAccepts = (process.env.NODE_MANY_ACCEPTS &&
                              process.env.NODE_MANY_ACCEPTS !== '0');
     }
