@@ -48,7 +48,8 @@ var path = require('path');
 var fs = require('fs');
 var rl = require('readline');
 var Console = require('console').Console;
-var EventEmitter = require('events').EventEmitter;
+var domain = require('domain');
+var debug = util.debuglog('repl');
 
 // If obj.hasOwnProperty has been overridden, then calling
 // obj.hasOwnProperty(prop) will break.
@@ -71,7 +72,7 @@ exports.writer = util.inspect;
 exports._builtinLibs = ['assert', 'buffer', 'child_process', 'cluster',
   'crypto', 'dgram', 'dns', 'domain', 'events', 'fs', 'http', 'https', 'net',
   'os', 'path', 'punycode', 'querystring', 'readline', 'stream',
-  'string_decoder', 'tls', 'tty', 'url', 'util', 'vm', 'zlib'];
+  'string_decoder', 'tls', 'tty', 'url', 'util', 'vm', 'zlib', 'smalloc'];
 
 
 function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
@@ -79,10 +80,8 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
     return new REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined);
   }
 
-  EventEmitter.call(this);
-
-  var options, input, output;
-  if (typeof prompt == 'object') {
+  var options, input, output, dom;
+  if (util.isObject(prompt)) {
     // an options object was given
     options = prompt;
     stream = options.stream || options.socket;
@@ -92,7 +91,8 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
     useGlobal = options.useGlobal;
     ignoreUndefined = options.ignoreUndefined;
     prompt = options.prompt;
-  } else if (typeof prompt != 'string') {
+    dom = options.domain;
+  } else if (!util.isString(prompt)) {
     throw new Error('An options Object, or a prompt String are required');
   } else {
     options = {};
@@ -100,28 +100,62 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
 
   var self = this;
 
+  self._domain = dom || domain.create();
+
   self.useGlobal = !!useGlobal;
   self.ignoreUndefined = !!ignoreUndefined;
 
-  self.eval = eval_ || function(code, context, file, cb) {
+  // just for backwards compat, see github.com/joyent/node/pull/7127
+  self.rli = this;
+
+  eval_ = eval_ || defaultEval;
+
+  function defaultEval(code, context, file, cb) {
     var err, result;
+    // first, create the Script object to check the syntax
     try {
-      if (self.useGlobal) {
-        result = vm.runInThisContext(code, file);
-      } else {
-        result = vm.runInContext(code, context, file);
-      }
+      var script = vm.createScript(code, {
+        filename: file,
+        displayErrors: false
+      });
     } catch (e) {
-      err = e;
+      debug('parse error %j', code, e);
+      if (isRecoverableError(e))
+        err = new Recoverable(e);
+      else
+        err = e;
     }
-    if (err && process.domain) {
-      process.domain.emit('error', err);
-      process.domain.exit();
+
+    if (!err) {
+      try {
+        if (self.useGlobal) {
+          result = script.runInThisContext({ displayErrors: false });
+        } else {
+          result = script.runInContext(context, { displayErrors: false });
+        }
+      } catch (e) {
+        err = e;
+        if (err && process.domain) {
+          debug('not recoverable, send to domain');
+          process.domain.emit('error', err);
+          process.domain.exit();
+          return;
+        }
+      }
     }
-    else {
-      cb(err, result);
-    }
-  };
+
+    cb(err, result);
+  }
+
+  self.eval = self._domain.bind(eval_);
+
+  self._domain.on('error', function(e) {
+    debug('domain error');
+    self.outputStream.write((e.stack || e) + '\n');
+    self.bufferedCommand = '';
+    self.lines.level = [];
+    self.displayPrompt();
+  });
 
   if (!input && !output) {
     // legacy API, passing a 'stream'/'socket' option
@@ -145,20 +179,20 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
 
   self.resetContext();
   self.bufferedCommand = '';
-
-  self.prompt = (prompt != undefined ? prompt : '> ');
+  self.lines.level = [];
 
   function complete(text, callback) {
     self.complete(text, callback);
   }
 
-  var rli = rl.createInterface({
-    input: self.inputStream,
-    output: self.outputStream,
-    completer: complete,
-    terminal: options.terminal
-  });
-  self.rli = rli;
+  rl.Interface.apply(this, [
+    self.inputStream,
+    self.outputStream,
+    complete,
+    options.terminal
+  ]);
+
+  self.setPrompt(!util.isUndefined(prompt) ? prompt : '> ');
 
   this.commands = {};
   defineDefaultCommands(this);
@@ -166,8 +200,8 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
   // figure out which "writer" function to use
   self.writer = options.writer || exports.writer;
 
-  if (typeof options.useColors === 'undefined') {
-    options.useColors = rli.terminal;
+  if (util.isUndefined(options.useColors)) {
+    options.useColors = self.terminal;
   }
   self.useColors = !!options.useColors;
 
@@ -178,34 +212,36 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
     };
   }
 
-  rli.setPrompt(self.prompt);
+  self.setPrompt(self._prompt);
 
-  rli.on('close', function() {
+  self.on('close', function() {
     self.emit('exit');
   });
 
   var sawSIGINT = false;
-  rli.on('SIGINT', function() {
-    var empty = rli.line.length === 0;
-    rli.clearLine();
+  self.on('SIGINT', function() {
+    var empty = self.line.length === 0;
+    self.clearLine();
 
     if (!(self.bufferedCommand && self.bufferedCommand.length > 0) && empty) {
       if (sawSIGINT) {
-        rli.close();
+        self.close();
         sawSIGINT = false;
         return;
       }
-      rli.output.write('(^C again to quit)\n');
+      self.output.write('(^C again to quit)\n');
       sawSIGINT = true;
     } else {
       sawSIGINT = false;
     }
 
     self.bufferedCommand = '';
+    self.lines.level = [];
     self.displayPrompt();
   });
 
-  rli.on('line', function(cmd) {
+  self.on('line', function(cmd) {
+    debug('line %j', cmd);
     sawSIGINT = false;
     var skipCatchall = false;
     cmd = trimWhitespace(cmd);
@@ -213,7 +249,7 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
     // Check to see if a REPL keyword was used. If it returns true,
     // display next prompt and return.
     if (cmd && cmd.charAt(0) === '.' && isNaN(parseFloat(cmd))) {
-      var matches = cmd.match(/^(\.[^\s]+)\s*(.*)$/);
+      var matches = cmd.match(/^\.([^\s]+)\s*(.*)$/);
       var keyword = matches && matches[1];
       var rest = matches && matches[2];
       if (self.parseREPLKeyword(keyword, rest) === true) {
@@ -225,68 +261,59 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
     }
 
     if (!skipCatchall) {
-      var evalCmd = self.bufferedCommand + cmd + '\n';
-
-      // This try is for determining if the command is complete, or should
-      // continue onto the next line.
-      // We try to evaluate both expressions e.g.
-      //  '{ a : 1 }'
-      // and statements e.g.
-      //  'for (var i = 0; i < 10; i++) console.log(i);'
-
-      // First we attempt to eval as expression with parens.
-      // This catches '{a : 1}' properly.
-      self.eval('(' + evalCmd + ')',
-                self.context,
-                'repl',
-                function(e, ret) {
-            if (e && !isSyntaxError(e)) return finish(e);
-
-            if (typeof ret === 'function' &&
-                /^[\r\n\s]*function/.test(evalCmd) ||
-                e) {
-              // Now as statement without parens.
-              self.eval(evalCmd, self.context, 'repl', finish);
-            } else {
-              finish(null, ret);
-            }
-          });
+      var evalCmd = self.bufferedCommand + cmd;
+      if (/^\s*\{/.test(evalCmd) && /\}\s*$/.test(evalCmd)) {
+        // It's confusing for `{ a : 1 }` to be interpreted as a block
+        // statement rather than an object literal.  So, we first try
+        // to wrap it in parentheses, so that it will be interpreted as
+        // an expression.
+        evalCmd = '(' + evalCmd + ')\n';
+      } else {
+        // otherwise we just append a \n so that it will be either
+        // terminated, or continued onto the next expression if it's an
+        // unexpected end of input.
+        evalCmd = evalCmd + '\n';
+      }
 
+      debug('eval %j', evalCmd);
+      self.eval(evalCmd, self.context, 'repl', finish);
     } else {
       finish(null);
     }
 
     function finish(e, ret) {
-
+      debug('finish', e, ret);
       self.memory(cmd);
 
-      // If error was SyntaxError and not JSON.parse error
-      if (isSyntaxError(e)) {
-        if (!self.bufferedCommand && cmd.trim().match(/^npm /)) {
-          self.outputStream.write('npm should be run outside of the ' +
-                                  'node repl, in your normal shell.\n' +
-                                  '(Press Control-D to exit.)\n');
-          self.bufferedCommand = '';
-          self.displayPrompt();
-          return;
-        }
-
-        // Start buffering data like that:
-        // {
-        // ...  x: 1
-        // ... }
-        self.bufferedCommand += cmd + '\n';
+      if (e && !self.bufferedCommand && cmd.trim().match(/^npm /)) {
+        self.outputStream.write('npm should be run outside of the ' +
+                                'node repl, in your normal shell.\n' +
+                                '(Press Control-D to exit.)\n');
+        self.bufferedCommand = '';
         self.displayPrompt();
         return;
-      } else if (e) {
-        self.outputStream.write((e.stack || e) + '\n');
+      }
+
+      // If error was SyntaxError and not JSON.parse error
+      if (e) {
+        if (e instanceof Recoverable) {
+          // Start buffering data like that:
+          // {
+          // ...  x: 1
+          // ... }
+          self.bufferedCommand += cmd + '\n';
+          self.displayPrompt();
+          return;
+        } else {
+          self._domain.emit('error', e);
+        }
       }
 
       // Clear buffer if no SyntaxErrors
       self.bufferedCommand = '';
 
       // If we got any output - print it (if no error)
-      if (!e && (!self.ignoreUndefined || ret !== undefined)) {
+      if (!e && (!self.ignoreUndefined || !util.isUndefined(ret))) {
         self.context._ = ret;
         self.outputStream.write(self.writer(ret) + '\n');
       }
@@ -296,13 +323,13 @@ function REPLServer(prompt, stream, eval_, useGlobal, ignoreUndefined) {
     };
   });
 
-  rli.on('SIGCONT', function() {
+  self.on('SIGCONT', function() {
     self.displayPrompt(true);
   });
 
   self.displayPrompt();
 }
-inherits(REPLServer, EventEmitter);
+inherits(REPLServer, rl.Interface);
 exports.REPLServer = REPLServer;
 
 
@@ -314,7 +341,6 @@ exports.start = function(prompt, source, eval_, useGlobal, ignoreUndefined) {
   return repl;
 };
 
-
 REPLServer.prototype.createContext = function() {
   var context;
   if (this.useGlobal) {
@@ -356,19 +382,29 @@ REPLServer.prototype.createContext = function() {
 
 REPLServer.prototype.resetContext = function() {
   this.context = this.createContext();
+
+  // Allow REPL extensions to extend the new context
+  this.emit('reset', this.context);
 };
 
 REPLServer.prototype.displayPrompt = function(preserveCursor) {
-  var prompt = this.prompt;
+  var prompt = this._initialPrompt;
   if (this.bufferedCommand.length) {
     prompt = '...';
     var levelInd = new Array(this.lines.level.length).join('..');
     prompt += levelInd + ' ';
   }
-  this.rli.setPrompt(prompt);
-  this.rli.prompt(preserveCursor);
+
+  // Do not overwrite `_initialPrompt` here
+  REPLServer.super_.prototype.setPrompt.call(this, prompt);
+  this.prompt(preserveCursor);
 };
 
+// When invoked as an API method, overwrite _initialPrompt
+REPLServer.prototype.setPrompt = function setPrompt(prompt) {
+  this._initialPrompt = prompt;
+  REPLServer.super_.prototype.setPrompt.call(this, prompt);
+};
 
 // A stream to push an array into a REPL
 // used in REPLServer.complete
@@ -405,7 +441,7 @@ var simpleExpressionRE =
 // getter code.
 REPLServer.prototype.complete = function(line, callback) {
   // There may be local variables to evaluate, try a nested REPL
-  if (this.bufferedCommand != undefined && this.bufferedCommand.length) {
+  if (!util.isUndefined(this.bufferedCommand) && this.bufferedCommand.length) {
     // Get a new array of inputed lines
     var tmp = this.lines.slice();
     // Kill off all function declarations to push all local variables into
@@ -431,7 +467,7 @@ REPLServer.prototype.complete = function(line, callback) {
   // list of completion lists, one for each inheritance "level"
   var completionGroups = [];
 
-  var completeOn, match, filter, i, j, group, c;
+  var completeOn, match, filter, i, group, c;
 
   // REPL commands (e.g. ".break").
   var match = null;
@@ -529,7 +565,7 @@ REPLServer.prototype.complete = function(line, callback) {
       }
 
       // Resolve expr and get its completions.
-      var obj, memberGroups = [];
+      var memberGroups = [];
       if (!expr) {
         // If context is instance of vm.ScriptContext
         // Get global vars synchronously
@@ -547,7 +583,7 @@ REPLServer.prototype.complete = function(line, callback) {
           this.eval('.scope', this.context, 'repl', function(err, globals) {
             if (err || !globals) {
               addStandardGlobals(completionGroups, filter);
-            } else if (Array.isArray(globals[0])) {
+            } else if (util.isArray(globals[0])) {
               // Add grouped globals
               globals.forEach(function(group) {
                 completionGroups.push(group);
@@ -564,19 +600,19 @@ REPLServer.prototype.complete = function(line, callback) {
           // if (e) console.log(e);
 
           if (obj != null) {
-            if (typeof obj === 'object' || typeof obj === 'function') {
+            if (util.isObject(obj) || util.isFunction(obj)) {
               memberGroups.push(Object.getOwnPropertyNames(obj));
             }
             // works for non-objects
             try {
               var sentinel = 5;
               var p;
-              if (typeof obj === 'object' || typeof obj === 'function') {
+              if (util.isObject(obj) || util.isFunction(obj)) {
                 p = Object.getPrototypeOf(obj);
               } else {
                 p = obj.constructor ? obj.constructor.prototype : null;
               }
-              while (p !== null) {
+              while (!util.isNull(p)) {
                 memberGroups.push(Object.getOwnPropertyNames(p));
                 p = Object.getPrototypeOf(p);
                 // Circular refs possible? Let's guard against that.
@@ -641,7 +677,7 @@ REPLServer.prototype.complete = function(line, callback) {
         group.sort();
         for (var j = 0; j < group.length; j++) {
           c = group[j];
-          if (!hasOwnProperty(c)) {
+          if (!hasOwnProperty(uniq, c)) {
             completions.push(c);
             uniq[c] = true;
           }
@@ -675,12 +711,12 @@ REPLServer.prototype.parseREPLKeyword = function(keyword, rest) {
 
 
 REPLServer.prototype.defineCommand = function(keyword, cmd) {
-  if (typeof cmd === 'function') {
+  if (util.isFunction(cmd)) {
     cmd = {action: cmd};
-  } else if (typeof cmd.action !== 'function') {
+  } else if (!util.isFunction(cmd.action)) {
     throw new Error('bad argument, action must be a function');
   }
-  this.commands['.' + keyword] = cmd;
+  this.commands[keyword] = cmd;
 };
 
 REPLServer.prototype.memory = function memory(cmd) {
@@ -809,7 +845,7 @@ function defineDefaultCommands(repl) {
   repl.defineCommand('exit', {
     help: 'Exit the repl',
     action: function() {
-      this.rli.close();
+      this.close();
     }
   });
 
@@ -850,7 +886,7 @@ function defineDefaultCommands(repl) {
           this.displayPrompt();
           lines.forEach(function(line) {
             if (line) {
-              self.rli.write(line + '\n');
+              self.write(line + '\n');
             }
           });
         }
@@ -908,20 +944,15 @@ REPLServer.prototype.convertToContext = function(cmd) {
 };
 
 
-/**
- * Returns `true` if "e" is a SyntaxError, `false` otherwise.
- * This function filters out false positives likes JSON.parse() errors and
- * RegExp syntax errors.
- */
-function isSyntaxError(e) {
-  // Convert error to string
-  e = e && (e.stack || e.toString());
-  return e && e.match(/^SyntaxError/) &&
-      // RegExp syntax error
-      !e.match(/^SyntaxError: Invalid regular expression/) &&
-      !e.match(/^SyntaxError: Invalid flags supplied to RegExp constructor/) &&
-      // "strict mode" syntax errors
-      !e.match(/^SyntaxError: .*strict mode.*/i) &&
-      // JSON.parse() error
-      !e.match(/\n {4}at Object.parse \(native\)\n/);
+// If the error is that we've unexpectedly ended the input,
+// then let the user try to recover by adding more input.
+function isRecoverableError(e) {
+  return e &&
+      e.name === 'SyntaxError' &&
+      /^(Unexpected end of input|Unexpected token :)/.test(e.message);
 }
+
+function Recoverable(err) {
+  this.err = err;
+}
+inherits(Recoverable, SyntaxError);
