@@ -23,15 +23,13 @@ var Timer = process.binding('timer_wrap').Timer;
 var L = require('_linklist');
 var assert = require('assert').ok;
 
+var kOnTimeout = Timer.kOnTimeout | 0;
+
 // Timeout values > TIMEOUT_MAX are set to 1.
 var TIMEOUT_MAX = 2147483647; // 2^31-1
 
-var debug;
-if (process.env.NODE_DEBUG && /timer/.test(process.env.NODE_DEBUG)) {
-  debug = function() { require('util').error.apply(this, arguments); };
-} else {
-  debug = function() { };
-}
+var util = require('util');
+var debug = util.debuglog('timer');
 
 
 // IDLE TIMEOUTS
@@ -50,9 +48,7 @@ var lists = {};
 // the main function - creates lists on demand and the watchers associated
 // with them.
 function insert(item, msecs) {
-  item._idleStart = Date.now();
-  item._monotonicStartTime = Timer.now();
-
+  item._idleStart = Timer.now();
   item._idleTimeout = msecs;
 
   if (msecs < 0) return;
@@ -69,7 +65,7 @@ function insert(item, msecs) {
 
     lists[msecs] = list;
     list.msecs = msecs;
-    list.ontimeout = listOnTimeout;
+    list[kOnTimeout] = listOnTimeout;
   }
 
   L.append(list, item);
@@ -80,17 +76,17 @@ function listOnTimeout() {
   var msecs = this.msecs;
   var list = this;
 
-  debug('timeout callback ' + msecs);
+  debug('timeout callback %d', msecs);
 
   var now = Timer.now();
-  debug('now: ' + now);
+  debug('now: %s', now);
 
-  var first;
+  var diff, first, threw;
   while (first = L.peek(list)) {
-    var diff = now - first._monotonicStartTime;
+    diff = now - first._idleStart;
     if (diff < msecs) {
       list.start(msecs - diff, 0);
-      debug(msecs + ' list wait because diff is ' + diff);
+      debug('%d list wait because diff is %d', msecs, diff);
       return;
     } else {
       L.remove(first);
@@ -104,11 +100,13 @@ function listOnTimeout() {
       //
       // https://github.com/joyent/node/issues/2631
       var domain = first.domain;
-      if (domain && domain._disposed) continue;
+      if (domain && domain._disposed)
+        continue;
+
       try {
         if (domain)
           domain.enter();
-        var threw = true;
+        threw = true;
         first._onTimeout();
         if (domain)
           domain.exit();
@@ -121,7 +119,7 @@ function listOnTimeout() {
           var oldDomain = process.domain;
           process.domain = null;
           process.nextTick(function() {
-            list.ontimeout();
+            list[kOnTimeout]();
           });
           process.domain = oldDomain;
         }
@@ -129,7 +127,7 @@ function listOnTimeout() {
     }
   }
 
-  debug(msecs + ' list empty');
+  debug('%d list empty', msecs);
   assert(L.isEmpty(list));
   list.close();
   delete lists[msecs];
@@ -154,13 +152,21 @@ var unenroll = exports.unenroll = function(item) {
 
 // Does not start the time, just sets up the members needed.
 exports.enroll = function(item, msecs) {
+  if (!util.isNumber(msecs)) {
+    throw new TypeError('msecs must be a number');
+  }
+
+  if (msecs < 0 || !isFinite(msecs)) {
+    throw new RangeError('msecs must be a non-negative finite number');
+  }
+
   // if this item was already in a list somewhere
   // then we should unenroll it from that
   if (item._idleNext) unenroll(item);
 
   // Ensure that msecs fits into signed int32
-  if (msecs > 0x7fffffff) {
-    msecs = 0x7fffffff;
+  if (msecs > TIMEOUT_MAX) {
+    msecs = TIMEOUT_MAX;
   }
 
   item._idleTimeout = msecs;
@@ -173,13 +179,11 @@ exports.enroll = function(item, msecs) {
 exports.active = function(item) {
   var msecs = item._idleTimeout;
   if (msecs >= 0) {
-
     var list = lists[msecs];
     if (!list || L.isEmpty(list)) {
       insert(item, msecs);
     } else {
-      item._idleStart = Date.now();
-      item._monotonicStartTime = Timer.now();
+      item._idleStart = Timer.now();
       L.append(list, item);
     }
   }
@@ -229,8 +233,8 @@ exports.setTimeout = function(callback, after) {
 
 
 exports.clearTimeout = function(timer) {
-  if (timer && (timer.ontimeout || timer._onTimeout)) {
-    timer.ontimeout = timer._onTimeout = null;
+  if (timer && (timer[kOnTimeout] || timer._onTimeout)) {
+    timer[kOnTimeout] = timer._onTimeout = null;
     if (timer instanceof Timeout) {
       timer.close(); // for after === 0
     } else {
@@ -285,25 +289,28 @@ var Timeout = function(after) {
   this._idlePrev = this;
   this._idleNext = this;
   this._idleStart = null;
-  this._monotonicStartTime = null;
   this._onTimeout = null;
   this._repeat = false;
 };
 
+
+function unrefdHandle() {
+  this.owner._onTimeout();
+  if (!this.owner._repeat)
+    this.owner.close();
+}
+
+
 Timeout.prototype.unref = function() {
   if (!this._handle) {
-
-    var nowMonotonic = Timer.now();
-    if (!this._idleStart || !this._monotonicStartTime) {
-      this._idleStart = Date.now();
-      this._monotonicStartTime = nowMonotonic;
-    }
-
-    var delay = this._monotonicStartTime + this._idleTimeout - nowMonotonic;
+    var now = Timer.now();
+    if (!this._idleStart) this._idleStart = now;
+    var delay = this._idleStart + this._idleTimeout - now;
     if (delay < 0) delay = 0;
     exports.unenroll(this);
     this._handle = new Timer();
-    this._handle.ontimeout = this._onTimeout;
+    this._handle.owner = this;
+    this._handle[kOnTimeout] = unrefdHandle;
     this._handle.start(delay, 0);
     this._handle.domain = this.domain;
     this._handle.unref();
@@ -320,7 +327,7 @@ Timeout.prototype.ref = function() {
 Timeout.prototype.close = function() {
   this._onTimeout = null;
   if (this._handle) {
-    this._handle.ontimeout = null;
+    this._handle[kOnTimeout] = null;
     this._handle.close();
   } else {
     exports.unenroll(this);
@@ -333,31 +340,70 @@ L.init(immediateQueue);
 
 
 function processImmediate() {
-  var immediate = L.shift(immediateQueue);
+  var queue = immediateQueue;
+  var domain, immediate;
 
-  if (L.isEmpty(immediateQueue)) {
-    process._needImmediateCallback = false;
+  immediateQueue = {};
+  L.init(immediateQueue);
+
+  while (L.isEmpty(queue) === false) {
+    immediate = L.shift(queue);
+    domain = immediate.domain;
+
+    if (domain)
+      domain.enter();
+
+    var threw = true;
+    try {
+      immediate._onImmediate();
+      threw = false;
+    } finally {
+      if (threw) {
+        if (!L.isEmpty(queue)) {
+          // Handle any remaining on next tick, assuming we're still
+          // alive to do so.
+          while (!L.isEmpty(immediateQueue)) {
+            L.append(queue, L.shift(immediateQueue));
+          }
+          immediateQueue = queue;
+          process.nextTick(processImmediate);
+        }
+      }
+    }
+
+    if (domain)
+      domain.exit();
   }
 
-  if (immediate._onImmediate) {
-    if (immediate.domain) immediate.domain.enter();
-
-    immediate._onImmediate();
-
-    if (immediate.domain) immediate.domain.exit();
+  // Only round-trip to C++ land if we have to. Calling clearImmediate() on an
+  // immediate that's in |queue| is okay. Worst case is we make a superfluous
+  // call to NeedImmediateCallbackSetter().
+  if (L.isEmpty(immediateQueue)) {
+    process._needImmediateCallback = false;
   }
 }
 
 
+function Immediate() { }
+
+Immediate.prototype.domain = undefined;
+Immediate.prototype._onImmediate = undefined;
+Immediate.prototype._idleNext = undefined;
+Immediate.prototype._idlePrev = undefined;
+
+
 exports.setImmediate = function(callback) {
-  var immediate = {}, args;
+  var immediate = new Immediate();
+  var args, index;
 
   L.init(immediate);
 
   immediate._onImmediate = callback;
 
   if (arguments.length > 1) {
-    args = Array.prototype.slice.call(arguments, 1);
+    args = [];
+    for (index = 1; index < arguments.length; index++)
+      args.push(arguments[index]);
 
     immediate._onImmediate = function() {
       callback.apply(immediate, args);
@@ -369,7 +415,8 @@ exports.setImmediate = function(callback) {
     process._immediateCallback = processImmediate;
   }
 
-  if (process.domain) immediate.domain = process.domain;
+  if (process.domain)
+    immediate.domain = process.domain;
 
   L.append(immediateQueue, immediate);
 
@@ -390,50 +437,109 @@ exports.clearImmediate = function(immediate) {
 };
 
 
-// Internal APIs that need timeouts should use timers._unrefActive isntead of
+// Internal APIs that need timeouts should use timers._unrefActive instead of
 // timers.active as internal timeouts shouldn't hold the loop open
 
 var unrefList, unrefTimer;
 
+function _makeTimerTimeout(timer) {
+  var domain = timer.domain;
+  var msecs = timer._idleTimeout;
+
+  // Timer has been unenrolled by another timer that fired at the same time,
+  // so don't make it timeout.
+  if (!msecs || msecs < 0)
+    return;
+
+  if (!timer._onTimeout)
+    return;
+
+  if (domain && domain._disposed)
+    return;
+
+  try {
+    var threw = true;
+
+    if (domain) domain.enter();
+
+    debug('unreftimer firing timeout');
+    L.remove(timer);
+    timer._onTimeout();
+
+    threw = false;
+
+    if (domain)
+      domain.exit();
+  } finally {
+    if (threw) process.nextTick(unrefTimeout);
+  }
+}
 
 function unrefTimeout() {
   var now = Timer.now();
 
   debug('unrefTimer fired');
 
-  var first;
-  while (first = L.peek(unrefList)) {
-    var diff = now - first._monotonicStartTime;
+  var timeSinceLastActive;
+  var nextTimeoutTime;
+  var nextTimeoutDuration;
+  var minNextTimeoutTime;
+  var timersToTimeout = [];
 
-    if (diff < first._idleTimeout) {
-      diff = first._idleTimeout - diff;
-      unrefTimer.start(diff, 0);
-      unrefTimer.when = now + diff;
-      debug('unrefTimer rescheudling for later');
-      return;
+  // The actual timer fired and has not yet been rearmed,
+  // let's consider its next firing time is invalid for now.
+  // It may be set to a relevant time in the future once
+  // we scanned through the whole list of timeouts and if
+  // we find a timeout that needs to expire.
+  unrefTimer.when = -1;
+
+  // Iterate over the list of timeouts,
+  // call the onTimeout callback for those expired,
+  // and rearm the actual timer if the next timeout to expire
+  // will expire before the current actual timer.
+  var cur = unrefList._idlePrev;
+  while (cur != unrefList) {
+    timeSinceLastActive = now - cur._idleStart;
+
+    if (timeSinceLastActive < cur._idleTimeout) {
+      // This timer hasn't expired yet, but check if its expiring time is
+      // earlier than the actual timer's expiring time
+
+      nextTimeoutDuration = cur._idleTimeout - timeSinceLastActive;
+      nextTimeoutTime = now + nextTimeoutDuration;
+      if (minNextTimeoutTime == null ||
+          (nextTimeoutTime < minNextTimeoutTime)) {
+        // We found a timeout that will expire earlier,
+        // store its next timeout time now so that we
+        // can rearm the actual timer accordingly when
+        // we scanned through the whole list.
+        minNextTimeoutTime = nextTimeoutTime;
+      }
+    } else {
+      // We found a timer that expired. Do not call its _onTimeout callback
+      // right now, as it could mutate any item of the list (including itself).
+      // Instead, add it to another list that will be processed once the list
+      // of current timers has been fully traversed.
+      timersToTimeout.push(cur);
     }
 
-    L.remove(first);
-
-    var domain = first.domain;
-
-    if (!first._onTimeout) continue;
-    if (domain && domain._disposed) continue;
-
-    try {
-      if (domain) domain.enter();
-      var threw = true;
-      debug('unreftimer firing timeout');
-      first._onTimeout();
-      threw = false;
-      if (domain) domain.exit();
-    } finally {
-      if (threw) process.nextTick(unrefTimeout);
-    }
+    cur = cur._idlePrev;
   }
 
-  debug('unrefList is empty');
-  unrefTimer.when = -1;
+  var nbTimersToTimeout = timersToTimeout.length;
+  for (var timerIdx = 0; timerIdx < nbTimersToTimeout; ++timerIdx)
+    _makeTimerTimeout(timersToTimeout[timerIdx]);
+
+
+  // Rearm the actual timer with the timeout delay
+  // of the earliest timeout found.
+  if (minNextTimeoutTime != null) {
+    unrefTimer.start(minNextTimeoutTime - now, 0);
+    unrefTimer.when = minNextTimeoutTime;
+    debug('unrefTimer rescheduled');
+  } else if (L.isEmpty(unrefList)) {
+    debug('unrefList is empty');
+  }
 }
 
 
@@ -453,47 +559,20 @@ exports._unrefActive = function(item) {
     unrefTimer = new Timer();
     unrefTimer.unref();
     unrefTimer.when = -1;
-    unrefTimer.ontimeout = unrefTimeout;
+    unrefTimer[kOnTimeout] = unrefTimeout;
   }
 
-  var nowDate = Date.now();
-  var nowMonotonicTimestamp = Timer.now();
+  var now = Timer.now();
+  item._idleStart = now;
 
-  item._idleStart = nowDate;
-  item._monotonicStartTime = nowMonotonicTimestamp;
-
-  if (L.isEmpty(unrefList)) {
-    debug('unrefList empty');
-    L.append(unrefList, item);
+  var when = now + msecs;
 
+  // If the actual timer is set to fire too late, or not set to fire at all,
+  // we need to make it fire earlier
+  if (unrefTimer.when === -1 || unrefTimer.when > when) {
     unrefTimer.start(msecs, 0);
-    unrefTimer.when = nowMonotonicTimestamp + msecs;
+    unrefTimer.when = when;
     debug('unrefTimer scheduled');
-    return;
-  }
-
-  var when = nowMonotonicTimestamp + msecs;
-
-  debug('unrefList find where we can insert');
-
-  var cur, them;
-
-  for (cur = unrefList._idlePrev; cur != unrefList; cur = cur._idlePrev) {
-    them = cur._monotonicStartTime + cur._idleTimeout;
-
-    if (when < them) {
-      debug('unrefList inserting into middle of list');
-
-      L.append(cur, item);
-
-      if (unrefTimer.when > when) {
-        debug('unrefTimer is scheduled to fire too late, reschedule');
-        unrefTimer.start(msecs, 0);
-        unrefTimer.when = when;
-      }
-
-      return;
-    }
   }
 
   debug('unrefList append to end');
