@@ -21,7 +21,7 @@
 
 var formatRegExp = /%[sdj%]/g;
 exports.format = function(f) {
-  if (typeof f !== 'string') {
+  if (!isString(f)) {
     var objects = [];
     for (var i = 0; i < arguments.length; i++) {
       objects.push(inspect(arguments[i]));
@@ -38,13 +38,18 @@ exports.format = function(f) {
     switch (x) {
       case '%s': return String(args[i++]);
       case '%d': return Number(args[i++]);
-      case '%j': return JSON.stringify(args[i++]);
+      case '%j':
+        try {
+          return JSON.stringify(args[i++]);
+        } catch (_) {
+          return '[Circular]';
+        }
       default:
         return x;
     }
   });
   for (var x = args[i]; i < len; x = args[++i]) {
-    if (x === null || typeof x !== 'object') {
+    if (isNull(x) || !isObject(x)) {
       str += ' ' + x;
     } else {
       str += ' ' + inspect(x);
@@ -58,6 +63,13 @@ exports.format = function(f) {
 // Returns a modified function which warns once by default.
 // If --no-deprecation is set, then it is a no-op.
 exports.deprecate = function(fn, msg) {
+  // Allow for deprecating things in the process of starting up.
+  if (isUndefined(global.process)) {
+    return function() {
+      return exports.deprecate(fn, msg).apply(this, arguments);
+    };
+  }
+
   if (process.noDeprecation === true) {
     return fn;
   }
@@ -65,7 +77,9 @@ exports.deprecate = function(fn, msg) {
   var warned = false;
   function deprecated() {
     if (!warned) {
-      if (process.traceDeprecation) {
+      if (process.throwDeprecation) {
+        throw new Error(msg);
+      } else if (process.traceDeprecation) {
         console.trace(msg);
       } else {
         console.error(msg);
@@ -79,29 +93,24 @@ exports.deprecate = function(fn, msg) {
 };
 
 
-exports.print = function() {
-  for (var i = 0, len = arguments.length; i < len; ++i) {
-    process.stdout.write(String(arguments[i]));
-  }
-};
-
-
-exports.puts = function() {
-  for (var i = 0, len = arguments.length; i < len; ++i) {
-    process.stdout.write(arguments[i] + '\n');
-  }
-};
-
-
-exports.debug = function(x) {
-  process.stderr.write('DEBUG: ' + x + '\n');
-};
-
-
-var error = exports.error = function(x) {
-  for (var i = 0, len = arguments.length; i < len; ++i) {
-    process.stderr.write(arguments[i] + '\n');
+var debugs = {};
+var debugEnviron;
+exports.debuglog = function(set) {
+  if (isUndefined(debugEnviron))
+    debugEnviron = process.env.NODE_DEBUG || '';
+  set = set.toUpperCase();
+  if (!debugs[set]) {
+    if (new RegExp('\\b' + set + '\\b', 'i').test(debugEnviron)) {
+      var pid = process.pid;
+      debugs[set] = function() {
+        var msg = exports.format.apply(exports, arguments);
+        console.error('%s %d: %s', set, pid, msg);
+      };
+    } else {
+      debugs[set] = function() {};
+    }
   }
+  return debugs[set];
 };
 
 
@@ -122,7 +131,7 @@ function inspect(obj, opts) {
   // legacy...
   if (arguments.length >= 3) ctx.depth = arguments[2];
   if (arguments.length >= 4) ctx.colors = arguments[3];
-  if (typeof opts === 'boolean') {
+  if (isBoolean(opts)) {
     // legacy...
     ctx.showHidden = opts;
   } else if (opts) {
@@ -130,10 +139,10 @@ function inspect(obj, opts) {
     exports._extend(ctx, opts);
   }
   // set default options
-  if (typeof ctx.showHidden === 'undefined') ctx.showHidden = false;
-  if (typeof ctx.depth === 'undefined') ctx.depth = 2;
-  if (typeof ctx.colors === 'undefined') ctx.colors = false;
-  if (typeof ctx.customInspect === 'undefined') ctx.customInspect = true;
+  if (isUndefined(ctx.showHidden)) ctx.showHidden = false;
+  if (isUndefined(ctx.depth)) ctx.depth = 2;
+  if (isUndefined(ctx.colors)) ctx.colors = false;
+  if (isUndefined(ctx.customInspect)) ctx.customInspect = true;
   if (ctx.colors) ctx.stylize = stylizeWithColor;
   return formatValue(ctx, obj, ctx.depth);
 }
@@ -165,6 +174,7 @@ inspect.styles = {
   'undefined': 'grey',
   'null': 'bold',
   'string': 'green',
+  'symbol': 'green',
   'date': 'magenta',
   // "name": intentionally not styling
   'regexp': 'red'
@@ -202,12 +212,18 @@ function arrayToHash(array) {
 function formatValue(ctx, value, recurseTimes) {
   // Provide a hook for user-specified inspect functions.
   // Check that value is an object with an inspect function on it
-  if (ctx.customInspect && value && typeof value.inspect === 'function' &&
+  if (ctx.customInspect &&
+      value &&
+      isFunction(value.inspect) &&
       // Filter out the util module, it's inspect function is special
       value.inspect !== exports.inspect &&
       // Also filter out any prototype objects using the circular check.
       !(value.constructor && value.constructor.prototype === value)) {
-    return String(value.inspect(recurseTimes));
+    var ret = value.inspect(recurseTimes, ctx);
+    if (!isString(ret)) {
+      ret = formatValue(ctx, ret, recurseTimes);
+    }
+    return ret;
   }
 
   // Primitive types cannot have properties
@@ -224,9 +240,31 @@ function formatValue(ctx, value, recurseTimes) {
     keys = Object.getOwnPropertyNames(value);
   }
 
+  // This could be a boxed primitive (new String(), etc.), check valueOf()
+  // NOTE: Avoid calling `valueOf` on `Date` instance because it will return
+  // a number which, when object has some additional user-stored `keys`,
+  // will be printed out.
+  var formatted;
+  var raw = value;
+  try {
+    // the .valueOf() call can fail for a multitude of reasons
+    if (!isDate(value))
+      raw = value.valueOf();
+  } catch (e) {
+    // ignore...
+  }
+
+  if (isString(raw)) {
+    // for boxed Strings, we have to remove the 0-n indexed entries,
+    // since they just noisey up the output and are redundant
+    keys = keys.filter(function(key) {
+      return !(key >= 0 && key < raw.length);
+    });
+  }
+
   // Some type of object without properties can be shortcutted.
   if (keys.length === 0) {
-    if (typeof value === 'function') {
+    if (isFunction(value)) {
       var name = value.name ? ': ' + value.name : '';
       return ctx.stylize('[Function' + name + ']', 'special');
     }
@@ -239,6 +277,19 @@ function formatValue(ctx, value, recurseTimes) {
     if (isError(value)) {
       return formatError(value);
     }
+    // now check the `raw` value to handle boxed primitives
+    if (isString(raw)) {
+      formatted = formatPrimitiveNoColor(ctx, raw);
+      return ctx.stylize('[String: ' + formatted + ']', 'string');
+    }
+    if (isNumber(raw)) {
+      formatted = formatPrimitiveNoColor(ctx, raw);
+      return ctx.stylize('[Number: ' + formatted + ']', 'number');
+    }
+    if (isBoolean(raw)) {
+      formatted = formatPrimitiveNoColor(ctx, raw);
+      return ctx.stylize('[Boolean: ' + formatted + ']', 'boolean');
+    }
   }
 
   var base = '', array = false, braces = ['{', '}'];
@@ -250,7 +301,7 @@ function formatValue(ctx, value, recurseTimes) {
   }
 
   // Make functions say that they are functions
-  if (typeof value === 'function') {
+  if (isFunction(value)) {
     var n = value.name ? ': ' + value.name : '';
     base = ' [Function' + n + ']';
   }
@@ -270,7 +321,25 @@ function formatValue(ctx, value, recurseTimes) {
     base = ' ' + formatError(value);
   }
 
-  if (keys.length === 0 && (!array || value.length == 0)) {
+  // Make boxed primitive Strings look like such
+  if (isString(raw)) {
+    formatted = formatPrimitiveNoColor(ctx, raw);
+    base = ' ' + '[String: ' + formatted + ']';
+  }
+
+  // Make boxed primitive Numbers look like such
+  if (isNumber(raw)) {
+    formatted = formatPrimitiveNoColor(ctx, raw);
+    base = ' ' + '[Number: ' + formatted + ']';
+  }
+
+  // Make boxed primitive Booleans look like such
+  if (isBoolean(raw)) {
+    formatted = formatPrimitiveNoColor(ctx, raw);
+    base = ' ' + '[Boolean: ' + formatted + ']';
+  }
+
+  if (keys.length === 0 && (!array || value.length === 0)) {
     return braces[0] + base + braces[1];
   }
 
@@ -300,26 +369,38 @@ function formatValue(ctx, value, recurseTimes) {
 
 
 function formatPrimitive(ctx, value) {
-  switch (typeof value) {
-    case 'undefined':
-      return ctx.stylize('undefined', 'undefined');
-
-    case 'string':
-      var simple = '\'' + JSON.stringify(value).replace(/^"|"$/g, '')
-                                               .replace(/'/g, "\\'")
-                                               .replace(/\\"/g, '"') + '\'';
-      return ctx.stylize(simple, 'string');
-
-    case 'number':
-      return ctx.stylize('' + value, 'number');
-
-    case 'boolean':
-      return ctx.stylize('' + value, 'boolean');
+  if (isUndefined(value))
+    return ctx.stylize('undefined', 'undefined');
+  if (isString(value)) {
+    var simple = '\'' + JSON.stringify(value).replace(/^"|"$/g, '')
+                                             .replace(/'/g, "\\'")
+                                             .replace(/\\"/g, '"') + '\'';
+    return ctx.stylize(simple, 'string');
   }
+  if (isNumber(value)) {
+    // Format -0 as '-0'. Strict equality won't distinguish 0 from -0,
+    // so instead we use the fact that 1 / -0 < 0 whereas 1 / 0 > 0 .
+    if (value === 0 && 1 / value < 0)
+      return ctx.stylize('-0', 'number');
+    return ctx.stylize('' + value, 'number');
+  }
+  if (isBoolean(value))
+    return ctx.stylize('' + value, 'boolean');
   // For some reason typeof null is "object", so special case here.
-  if (value === null) {
+  if (isNull(value))
     return ctx.stylize('null', 'null');
-  }
+  // es6 symbol primitive
+  if (isSymbol(value))
+    return ctx.stylize(value.toString(), 'symbol');
+}
+
+
+function formatPrimitiveNoColor(ctx, value) {
+  var stylize = ctx.stylize;
+  ctx.stylize = stylizeNoColor;
+  var str = formatPrimitive(ctx, value);
+  ctx.stylize = stylize;
+  return str;
 }
 
 
@@ -367,7 +448,7 @@ function formatProperty(ctx, value, recurseTimes, visibleKeys, key, array) {
   }
   if (!str) {
     if (ctx.seen.indexOf(desc.value) < 0) {
-      if (recurseTimes === null) {
+      if (isNull(recurseTimes)) {
         str = formatValue(ctx, desc.value, null);
       } else {
         str = formatValue(ctx, desc.value, recurseTimes - 1);
@@ -387,7 +468,7 @@ function formatProperty(ctx, value, recurseTimes, visibleKeys, key, array) {
       str = ctx.stylize('[Circular]', 'special');
     }
   }
-  if (typeof name === 'undefined') {
+  if (isUndefined(name)) {
     if (array && key.match(/^\d+$/)) {
       return str;
     }
@@ -398,7 +479,8 @@ function formatProperty(ctx, value, recurseTimes, visibleKeys, key, array) {
     } else {
       name = name.replace(/'/g, "\\'")
                  .replace(/\\"/g, '"')
-                 .replace(/(^"|"$)/g, "'");
+                 .replace(/(^"|"$)/g, "'")
+                 .replace(/\\\\/g, '\\');
       name = ctx.stylize(name, 'string');
     }
   }
@@ -408,11 +490,8 @@ function formatProperty(ctx, value, recurseTimes, visibleKeys, key, array) {
 
 
 function reduceToSingleString(output, base, braces) {
-  var numLinesEst = 0;
   var length = output.reduce(function(prev, cur) {
-    numLinesEst++;
-    if (cur.indexOf('\n') >= 0) numLinesEst++;
-    return prev + cur.length + 1;
+    return prev + cur.replace(/\u001b\[\d\d?m/g, '').length + 1;
   }, 0);
 
   if (length > 60) {
@@ -430,43 +509,89 @@ function reduceToSingleString(output, base, braces) {
 
 // NOTE: These type checking functions intentionally don't use `instanceof`
 // because it is fragile and can be easily faked with `Object.create()`.
-function isArray(ar) {
-  return Array.isArray(ar) ||
-         (typeof ar === 'object' && objectToString(ar) === '[object Array]');
-}
-exports.isArray = isArray;
+var isArray = exports.isArray = Array.isArray;
 
+function isBoolean(arg) {
+  return typeof arg === 'boolean';
+}
+exports.isBoolean = isBoolean;
+
+function isNull(arg) {
+  return arg === null;
+}
+exports.isNull = isNull;
+
+function isNullOrUndefined(arg) {
+  return arg == null;
+}
+exports.isNullOrUndefined = isNullOrUndefined;
+
+function isNumber(arg) {
+  return typeof arg === 'number';
+}
+exports.isNumber = isNumber;
+
+function isString(arg) {
+  return typeof arg === 'string';
+}
+exports.isString = isString;
+
+function isSymbol(arg) {
+  return typeof arg === 'symbol';
+}
+exports.isSymbol = isSymbol;
+
+function isUndefined(arg) {
+  return arg === void 0;
+}
+exports.isUndefined = isUndefined;
 
 function isRegExp(re) {
-  return typeof re === 'object' && objectToString(re) === '[object RegExp]';
+  return isObject(re) && objectToString(re) === '[object RegExp]';
 }
 exports.isRegExp = isRegExp;
 
+function isObject(arg) {
+  return typeof arg === 'object' && arg !== null;
+}
+exports.isObject = isObject;
 
 function isDate(d) {
-  return typeof d === 'object' && objectToString(d) === '[object Date]';
+  return isObject(d) && objectToString(d) === '[object Date]';
 }
 exports.isDate = isDate;
 
-
 function isError(e) {
-  return typeof e === 'object' && objectToString(e) === '[object Error]';
+  return isObject(e) &&
+      (objectToString(e) === '[object Error]' || e instanceof Error);
 }
 exports.isError = isError;
 
+function isFunction(arg) {
+  return typeof arg === 'function';
+}
+exports.isFunction = isFunction;
+
+function isPrimitive(arg) {
+  return arg === null ||
+         typeof arg === 'boolean' ||
+         typeof arg === 'number' ||
+         typeof arg === 'string' ||
+         typeof arg === 'symbol' ||  // ES6 symbol
+         typeof arg === 'undefined';
+}
+exports.isPrimitive = isPrimitive;
+
+function isBuffer(arg) {
+  return arg instanceof Buffer;
+}
+exports.isBuffer = isBuffer;
 
 function objectToString(o) {
   return Object.prototype.toString.call(o);
 }
 
 
-exports.p = exports.deprecate(function() {
-  for (var i = 0, len = arguments.length; i < len; ++i) {
-    error(exports.inspect(arguments[i]));
-  }
-}, 'util.p: Use console.error() instead.');
-
-
 function pad(n) {
   return n < 10 ? '0' + n.toString(10) : n.toString(10);
 }
@@ -485,17 +610,95 @@ function timestamp() {
 }
 
 
-exports.log = function(msg) {
-  exports.puts(timestamp() + ' - ' + msg.toString());
+// log is just a thin wrapper to console.log that prepends a timestamp
+exports.log = function() {
+  console.log('%s - %s', timestamp(), exports.format.apply(exports, arguments));
 };
 
 
+/**
+ * Inherit the prototype methods from one constructor into another.
+ *
+ * The Function.prototype.inherits from lang.js rewritten as a standalone
+ * function (not on Function.prototype). NOTE: If this file is to be loaded
+ * during bootstrapping this function needs to be rewritten using some native
+ * functions as prototype setup using normal JavaScript does not work as
+ * expected during bootstrapping (see mirror.js in r114903).
+ *
+ * @param {function} ctor Constructor function which needs to inherit the
+ *     prototype.
+ * @param {function} superCtor Constructor function to inherit prototype from.
+ */
+exports.inherits = function(ctor, superCtor) {
+  ctor.super_ = superCtor;
+  ctor.prototype = Object.create(superCtor.prototype, {
+    constructor: {
+      value: ctor,
+      enumerable: false,
+      writable: true,
+      configurable: true
+    }
+  });
+};
+
+exports._extend = function(origin, add) {
+  // Don't do anything if add isn't an object
+  if (!add || !isObject(add)) return origin;
+
+  var keys = Object.keys(add);
+  var i = keys.length;
+  while (i--) {
+    origin[keys[i]] = add[keys[i]];
+  }
+  return origin;
+};
+
+function hasOwnProperty(obj, prop) {
+  return Object.prototype.hasOwnProperty.call(obj, prop);
+}
+
+
+// Deprecated old stuff.
+
+exports.p = exports.deprecate(function() {
+  for (var i = 0, len = arguments.length; i < len; ++i) {
+    console.error(exports.inspect(arguments[i]));
+  }
+}, 'util.p: Use console.error() instead');
+
+
 exports.exec = exports.deprecate(function() {
   return require('child_process').exec.apply(this, arguments);
 }, 'util.exec is now called `child_process.exec`.');
 
 
-function pump(readStream, writeStream, callback) {
+exports.print = exports.deprecate(function() {
+  for (var i = 0, len = arguments.length; i < len; ++i) {
+    process.stdout.write(String(arguments[i]));
+  }
+}, 'util.print: Use console.log instead');
+
+
+exports.puts = exports.deprecate(function() {
+  for (var i = 0, len = arguments.length; i < len; ++i) {
+    process.stdout.write(arguments[i] + '\n');
+  }
+}, 'util.puts: Use console.log instead');
+
+
+exports.debug = exports.deprecate(function(x) {
+  process.stderr.write('DEBUG: ' + x + '\n');
+}, 'util.debug: Use console.error instead');
+
+
+exports.error = exports.deprecate(function(x) {
+  for (var i = 0, len = arguments.length; i < len; ++i) {
+    process.stderr.write(arguments[i] + '\n');
+  }
+}, 'util.error: Use console.error instead');
+
+
+exports.pump = exports.deprecate(function(readStream, writeStream, callback) {
   var callbackCalled = false;
 
   function call(a, b, c) {
@@ -530,48 +733,19 @@ function pump(readStream, writeStream, callback) {
     readStream.destroy();
     call(err);
   });
-}
-exports.pump = exports.deprecate(pump,
-    'util.pump() is deprecated. Use ReadableStream.prototype.pump() instead.');
+}, 'util.pump(): Use readableStream.pipe() instead');
 
 
-/**
- * Inherit the prototype methods from one constructor into another.
- *
- * The Function.prototype.inherits from lang.js rewritten as a standalone
- * function (not on Function.prototype). NOTE: If this file is to be loaded
- * during bootstrapping this function needs to be rewritten using some native
- * functions as prototype setup using normal JavaScript does not work as
- * expected during bootstrapping (see mirror.js in r114903).
- *
- * @param {function} ctor Constructor function which needs to inherit the
- *     prototype.
- * @param {function} superCtor Constructor function to inherit prototype from.
- */
-exports.inherits = function(ctor, superCtor) {
-  ctor.super_ = superCtor;
-  ctor.prototype = Object.create(superCtor.prototype, {
-    constructor: {
-      value: ctor,
-      enumerable: false,
-      writable: true,
-      configurable: true
-    }
-  });
+var uv;
+exports._errnoException = function(err, syscall, original) {
+  if (isUndefined(uv)) uv = process.binding('uv');
+  var errname = uv.errname(err);
+  var message = syscall + ' ' + errname;
+  if (original)
+    message += ' ' + original;
+  var e = new Error(message);
+  e.code = errname;
+  e.errno = errname;
+  e.syscall = syscall;
+  return e;
 };
-
-exports._extend = function(origin, add) {
-  // Don't do anything if add isn't an object
-  if (!add || typeof add !== 'object') return origin;
-
-  var keys = Object.keys(add);
-  var i = keys.length;
-  while (i--) {
-    origin[keys[i]] = add[keys[i]];
-  }
-  return origin;
-};
-
-function hasOwnProperty(obj, prop) {
-  return Object.prototype.hasOwnProperty.call(obj, prop);
-}
