@@ -19,15 +19,16 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+
+
+
 var common = require('../common');
 var assert = require('assert');
+var vm = require('vm');
 
-common.error('before');
-
-var Script = process.binding('evals').NodeScript;
+console.error('before');
 
 // undefined reference
-script = new Script('foo.bar = 5;');
-script.runInNewContext();
+vm.runInNewContext('foo.bar = 5;');
 
-common.error('after');
+console.error('after');
