@@ -19,13 +19,19 @@
 // OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
+
+
+
 var common = require('../common');
 var assert = require('assert');
 
 var path = require('path'),
     fs = require('fs'),
     fn = path.join(common.fixturesDir, 'non-existent'),
-    existingFile = path.join(common.fixturesDir, 'exit.js');
+    existingFile = path.join(common.fixturesDir, 'exit.js'),
+    existingFile2 = path.join(common.fixturesDir, 'create-file.js'),
+    existingDir = path.join(common.fixturesDir, 'empty'),
+    existingDir2 = path.join(common.fixturesDir, 'keys');
 
 // ASYNC_CALL
 
@@ -46,6 +52,14 @@ fs.link(fn, 'foo', function(err) {
   assert.ok(0 <= err.message.indexOf(fn));
 });
 
+fs.link(existingFile, existingFile2, function(err) {
+  assert.ok(0 <= err.message.indexOf(existingFile2));
+});
+
+fs.symlink(existingFile, existingFile2, function(err) {
+  assert.ok(0 <= err.message.indexOf(existingFile2));
+});
+
 fs.unlink(fn, function(err) {
   assert.ok(0 <= err.message.indexOf(fn));
 });
@@ -54,6 +68,10 @@ fs.rename(fn, 'foo', function(err) {
   assert.ok(0 <= err.message.indexOf(fn));
 });
 
+fs.rename(existingDir, existingDir2, function(err) {
+  assert.ok(0 <= err.message.indexOf(existingDir2));
+});
+
 fs.rmdir(fn, function(err) {
   assert.ok(0 <= err.message.indexOf(fn));
 });
@@ -131,6 +149,22 @@ try {
   assert.ok(0 <= err.message.indexOf(fn));
 }
 
+try {
+  ++expected;
+  fs.linkSync(existingFile, existingFile2);
+} catch (err) {
+  errors.push('link');
+  assert.ok(0 <= err.message.indexOf(existingFile2));
+}
+
+try {
+  ++expected;
+  fs.symlinkSync(existingFile, existingFile2);
+} catch (err) {
+  errors.push('symlink');
+  assert.ok(0 <= err.message.indexOf(existingFile2));
+}
+
 try {
   ++expected;
   fs.unlinkSync(fn);
@@ -171,6 +205,14 @@ try {
   assert.ok(0 <= err.message.indexOf(fn));
 }
 
+try {
+  ++expected;
+  fs.renameSync(existingDir, existingDir2);
+} catch (err) {
+  errors.push('rename');
+  assert.ok(0 <= err.message.indexOf(existingDir2));
+}
+
 try {
   ++expected;
   fs.readdirSync(fn);
@@ -179,7 +221,7 @@ try {
   assert.ok(0 <= err.message.indexOf(fn));
 }
 
-process.addListener('exit', function() {
+process.on('exit', function() {
   assert.equal(expected, errors.length,
                'Test fs sync exceptions raised, got ' + errors.length +
                ' expected ' + expected);
