@@ -20,7 +20,7 @@
 // USE OR OTHER DEALINGS IN THE SOFTWARE.
 
 if (!process.versions.openssl) {
-  console.error("Skipping because node compiled without OpenSSL.");
+  console.error('Skipping because node compiled without OpenSSL.');
   process.exit(0);
 }
 
@@ -31,14 +31,7 @@ var http = require('http');
 var cp = require('child_process');
 var fs = require('fs');
 
-var filename = require('path').join(common.tmpDir || '/tmp', 'big');
-
-// Clean up after ourselves. Leaving files around
-// in the tmp/ directory may break tests that depend
-// on a certain number of files being there.
-process.on('exit', function() {
-  fs.unlink(filename);
-});
+var filename = require('path').join(common.tmpDir, 'big');
 
 var count = 0;
 function maybeMakeRequest() {
@@ -50,13 +43,14 @@ function maybeMakeRequest() {
     var hex = stdout.match(/([A-Fa-f0-9]{40})/)[0];
     assert.equal('8c206a1a87599f532ce68675536f0b1546900d7a', hex);
     console.log('got the correct response');
+    fs.unlink(filename);
     server.close();
   });
 }
 
 
 var ddcmd = common.ddCommand(filename, 10240);
-console.log("dd command: ", ddcmd);
+console.log('dd command: ', ddcmd);
 
 cp.exec(ddcmd, function(err, stdout, stderr) {
   if (err) throw err;
@@ -79,7 +73,7 @@ var server = http.createServer(function(req, res) {
   cat.on('exit', function(code) {
     if (code !== 0) {
       console.error('subprocess exited with code ' + code);
-      exit(1);
+      process.exit(1);
     }
     res.end();
   });
