@@ -70,8 +70,11 @@ var gotEnd = false;
 var bodyBuffer = '';
 
 server.listen(common.PORT, function() {
-    console.log('1) Making Request');
-  var req = https.get({ port: common.PORT }, function(res) {
+  console.log('1) Making Request');
+  var req = https.get({
+    port: common.PORT,
+    rejectUnauthorized: false
+  }, function(res) {
     server.close();
     console.log('3) Client got response headers.');
 
